
import pandas as pd
from binascii import hexlify
from collections import namedtuple
from inspect import stack, getmodule
from json import loads
from random import choice
from re import sub
from string import ascii_letters

from ..stoats.names import BOOKMARK


class WarnDict(dict):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            msg = self.__msg % (item,)
            self.__log.debug(msg)
            raise KeyError(msg)


class WarnList(list):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except IndexError:
            msg = self.__msg % item
            self.__log.debug(msg)
            raise IndexError(msg)


def tohex(data):
    return hexlify(data).decode('ascii')


def assert_attr(instance, *attrs):
    for attr in attrs:
        if getattr(instance, attr) is None:
            raise Exception('No %s defined' % attr)


def kargs_to_attr(**kargs):
    return dict_to_attr(kargs)


def dict_to_attr(kargs):
    return namedtuple('Attr', kargs.keys(), rename=True)(*kargs.values())


class MutableAttr(dict):

    def __init__(self, *args, none=False, **kargs):
        self.__none = none
        super().__init__(*args, **kargs)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            if self.__none:
                return None
            else:
                raise AttributeError(name)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            self[name] = value

    def _to_dict(self):
        return self.__dict__


class MissingReference(Exception): pass


def reftuple(name, *args, **kargs):
    '''
    Like a namedtuple, but expands $ strings using a database session and date
    (# is similar, but also does JSON parsing).
    '''

    from ..squeal import StatisticJournal, StatisticName

    class klass(namedtuple(name, *args, **kargs)):

        def expand(self, log, s, time, owner=None, constraint=None):
            instance = self
            for name in self._fields:
                value = getattr(instance, name)
                if isinstance(value, str) and len(value) > 0 and value[0] in '$#':
                    log.info(f'Expanding {value} for {name}')
                    replacement = self._lookup(log, s, time, value[1:], value[0] == '#',
                                               default_value=self._fields_defaults.get(name, None),
                                               default_owner=owner, default_constraint=constraint)
                    instance = instance._replace(**{name: replacement})
            return instance

        def _lookup(self, log, s, time, name, json,
                    default_value=None, default_owner=None, default_constraint=None):
            if isinstance(name, str) and len(name) > 0 and name[0] in '$#':
                name = self._lookup(log, s, time, name[1:], name[0] == '#',
                                    default_owner=default_owner, default_constraint=default_constraint)
            owner, name, constraint = StatisticName.parse(name, default_owner=default_owner,
                                                          default_constraint=default_constraint)
            value = StatisticJournal.before(s, time, name, owner, constraint)
            if value is None:
                if default_value:
                    log.warning(f'No value found for {owner}:{name}:{constraint} (default {default_value})')
                    value = default_value
                else:
                    raise MissingReference(f'No value found for {owner}:{name}:{constraint} (and no default)')
            else:
                value = value.value
                if json:
                    log.debug(f'Unpacking JSON "{value}"')
                    value = loads(value)
                log.info(f'{name} -> {value}')
            return value

    klass.__name__ = name
    caller = stack()[1]
    klass.__module__ = getmodule(None, caller.filename).__name__
    return klass


class MaxDict(dict):

    def __init__(self, kv):
        super().__init__()
        for key, value in kv:
            if key in self:
                self[key] = max(value, self[key])
            else:
                self[key] = value


def tmp_name():
    return ''.join(choice(ascii_letters) for _ in range(8))


def nearest_index(df, name, value):
    exactmatch = df.loc[df[name] == value]
    if not exactmatch.empty:
        return exactmatch.index[0]
    else:
        lower = df.loc[df[name] < value].index.dropna()
        upper = df.loc[df[name] > value].index.dropna()
        if lower.empty:
            return upper.min()
        elif upper.empty:
            return lower.max()
        else:
            if abs(value - df.loc[lower.max()][name]) < abs(value - df.loc[upper.min()][name]):
                return lower.max()
            else:
                return upper.min()


def get_index_loc(df, value):
    loc = df.index.get_loc(value)
    try:
        return loc.start  # if slice, take first
    except AttributeError:
        return loc  # otherwise, simple value


def linscale(series, lo=0, hi=None, min=0, max=1, gamma=1):
    lo = series.dropna().min() if lo is None else lo
    hi = series.dropna().max() if hi is None else hi
    return (min + (max - min) * ((series - lo) / (hi - lo)) ** gamma).clip(min, max)


def sorted_numeric_labels(labels, text=None):
    if text:
        labels = [label for label in labels if text in label]
    return sorted(labels, key=lambda label: int(sub(r'\D', '', label)))


def interpolate_freq(df, freq, **kargs):
    left = pd.DataFrame(index=pd.date_range(df.index.min(), df.index.max(), freq=freq))
    return left_interpolate(left, df, **kargs)


def left_interpolate(left, right, **kargs):
    '''
    interpolate right so that it is on the same index as left.
    '''
    # neater solution https://stackoverflow.com/questions/47148446/pandas-resample-interpolate-is-producing-nans
    # option 1 with reindexing
    tmp = tmp_name()
    left[tmp] = True
    both = left.join(right, how='outer').interpolate(**kargs)
    return both.loc[both[tmp] == True].drop(columns=[tmp])


def bookend(df, column=BOOKMARK):
    # https://stackoverflow.com/questions/53927414/get-only-the-first-and-last-rows-of-each-group-with-pandas
    g = df.groupby(column)
    return pd.concat([g.head(1), g.tail(1)]).drop_duplicates().sort_index()




from sqlalchemy import inspect

from ..lib.data import dict_to_attr


class ORMUtils:

    def _get_or_create(self, session, cls, **kargs):
        query = session.query(cls)
        for (name, value) in kargs.items():
            query = query.filter(getattr(cls, name) == value)
        instance = query.one_or_none()
        if instance is None:
            instance = cls(**kargs)
            session.add(instance)
        return instance


def tables(*classes):
    return dict_to_attr(dict((cls.__name__, inspect(cls).local_table) for cls in classes))


def add(s, instance):
    s.add(instance)
    return instance




from collections import namedtuple
from logging import getLogger

log = getLogger(__name__)


def make_waypoint(names, extra=None):
    names = list(names)
    if extra:
        names += [extra]
    names = ['time'] + names
    defaults = [None] * len(names)
    return namedtuple('Waypoint', names, defaults=defaults)


def filter_none(names, waypoints):
    names = list(names)
    return [w for w in waypoints if all(n in w._fields and getattr(w, n) is not None for n in names)]



from logging import getLogger

from .records import restrict_names
from .tokens import State, FileHeader, token_factory, Checksum
from ..profile.profile import read_profile
from ...lib.data import tohex

log = getLogger(__name__)


def parse_data(data, types, messages, no_validate=False, max_delta_t=None):

    state = State(types, messages, max_delta_t=max_delta_t)

    def generator():
        offset = 0
        try:
            file_header = FileHeader(data[offset:])
            yield offset, file_header
            offset = len(file_header)
            file_header.validate(data, log, quiet=no_validate)
            while len(data) - offset > 2:
                token = token_factory(data[offset:], state)
                yield offset, token
                offset += len(token)
            checksum = Checksum(data[offset:])
            yield offset, checksum
            checksum.validate(data, log, quiet=no_validate)
        except Exception as e:
            log.warning('"%s" at offset %d' % (e, offset))
            dump(data, offset)
            raise

    return state, generator()


def dump(data, offset, rows=3, blocks=6, block=4):
    for row in range(rows):
        line = '%06d' % offset
        for b in range(blocks):
            line += ' '
            line += tohex(data[offset:offset+block])
            offset += block
        log.info(line)
        if offset >= len(data): return


def filtered_tokens(data,
                    after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
                    warn=False, no_validate=False, max_delta_t=None, profile_path=None):

    types, messages = read_profile(warn=warn, profile_path=profile_path)
    state, tokens = parse_data(data, types, messages, no_validate=no_validate, max_delta_t=max_delta_t)

    def generator():
        first_record = 0 if (after_records is None) else None
        first_bytes = 0 if (after_bytes is None) else None
        for i, (offset, token) in enumerate(tokens):
            if (first_record is None and (after_records is not None and i >= after_records)) or \
                    (first_bytes is None and (after_bytes is not None and offset >= after_bytes)):
                first_record = i
                first_bytes = offset
            if (first_record is not None and (limit_records < 0 or i - first_record < limit_records)) and \
                    (first_bytes is not None and (limit_bytes < 0 or offset - first_bytes < limit_bytes)):
                yield i, offset, token

    return types, messages, generator()


def filtered_records(data,
                     after_bytes=None, limit_bytes=-1, after_records=None, limit_records=-1,
                     record_names=None, field_names=None,
                     warn=False, no_validate=False, internal=False, max_delta_t=None,
                     profile_path=None, pipeline=None):

    if pipeline is None: pipeline = []
    if field_names: pipeline.append(restrict_names(field_names))
    types, messages = read_profile(warn=warn, profile_path=profile_path)
    state, tokens = parse_data(data, types, messages, no_validate=no_validate, max_delta_t=max_delta_t)

    def generator():
        first_record = 0 if (after_records is None) else None
        first_bytes = 0 if (after_bytes is None) else None
        for i, (offset, token) in enumerate(tokens):
            if (first_record is None and (after_records is not None and i >= after_records)) or \
                    (first_bytes is None and (after_bytes is not None and offset >= after_bytes)):
                first_record = i
                first_bytes = offset
            record = token.parse_token(warn=warn)
            if state.accumulators: record = record.force(*pipeline)
            if (internal or token.is_user) and (not record_names or record.name in record_names) and \
                    (first_record is not None and (limit_records < 0 or i - first_record < limit_records)) and \
                    (first_bytes is not None and (limit_bytes < 0 or offset - first_bytes < limit_bytes)):
                if not state.accumulators: record = record.force(*pipeline)
                yield i, offset, record

    return types, messages, generator()




from json import dumps

from .database import add_statistics, add_enum_constant, set_constant, name_constant
from ..squeal.types import short_cls
from ..stoats.calculate.impulse import HRImpulse, ImpulseCalculator
from ..stoats.calculate.response import Response, ResponseCalculator
from ..stoats.names import FITNESS_D, FATIGUE_D, HR_IMPULSE_10
from ..stoats.read.segment import SegmentReader


def add_impulse(s, c, activity_group, start=1e-6, fitness=((42, 1),), fatigue=((7, 5),)):
    '''
    Add configuration for a fitness/fatigue impulse model based on HR zones.

    This adds:
    * HeartRateCalculator pipeline class to calculate HR zone and impulse
      * Impulse constant with parameters for impulse calculation
    * ImpulseCalculator pipeline class to calculate fitness and fatigue
      * Fitness constant with parameters for fitness calculation
      * Fatigue constant with parameters for fatigue calculation
    '''

    activity_group_constraint = str(activity_group)
    responses = []

    hr_impulse_name = name_constant('HRImpulse', activity_group)
    hr_impulse = add_enum_constant(s, hr_impulse_name, HRImpulse, single=True,
                                   constraint=activity_group_constraint,
                                   description='Data needed to calculate the FF-model impulse from HR zones - ' +
                                               'see HRImpulse enum')
    # range 1-6, gamma 2 and decay of 84 days seems to fit my data best
    set_constant(s, hr_impulse, dumps({'dest_name': HR_IMPULSE_10, 'gamma': 2.0, 'zero': 1, 'one': 6, 'max_secs': 60}))

    for days, scale in fitness:
        name = FITNESS_D % days
        constant = name_constant(name, activity_group)
        responses.append(constant)
        fitness = add_enum_constant(s, constant, Response, single=True, constraint=activity_group_constraint,
                                    description=f'Data needed to calculate the FF-model fitness for {days}d - ' +
                                    'see Response enum')
        set_constant(s, fitness, dumps({'src_owner': short_cls(ImpulseCalculator),
                                        'dest_name': name, 'tau_days': days, 'scale': scale, 'start': start}))

    for days, scale in fatigue:
        name = FATIGUE_D % days
        constant = name_constant(name, activity_group)
        responses.append(constant)
        fitness = add_enum_constant(s, constant, Response, single=True, constraint=activity_group_constraint,
                                    description=f'Data needed to calculate the FF-model fatigue for {days}d - ' +
                                    'see Response enum')
        set_constant(s, fitness, dumps({'src_owner': short_cls(ImpulseCalculator),
                                        'dest_name': name, 'tau_days': days, 'scale': scale, 'start': start}))

    add_statistics(s, ImpulseCalculator, c, owner_in=short_cls(SegmentReader),
                   impulse_ref=hr_impulse_name)
    add_statistics(s, ResponseCalculator, c, owner_in=short_cls(ImpulseCalculator),
                   responses_ref=responses, impulse_ref=hr_impulse_name)




from bokeh.io import output_file
from bokeh.palettes import magma

from ch2.data import *
from ch2.data.plot.calendar import *
from ch2.lib import *
from ch2.stoats.names import _d
from ch2.uranus.decorator import template


@template
def calendar():

    '''
    # Calendar
    '''

    '''
    $contents
    '''

    s = session('-v5')
    output_file(filename='/dev/null')

    '''
    ## Distance

    Larger distances have larger symbols.
    
    Place the cursor over the symbol for more information.
    '''

    df = statistics(s, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB)
    df = coallesce(df, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB)
    df[DISTANCE_KM] = df[ACTIVE_DISTANCE] / 1000
    df['Duration'] = df[ACTIVE_TIME].map(format_seconds)
    if present(df, TOTAL_CLIMB):
        df.loc[df[TOTAL_CLIMB].isna(), [TOTAL_CLIMB]] = 0

    calendar = Calendar(df, title=DISTANCE, not_hover=[ACTIVE_DISTANCE, ACTIVE_TIME])
    calendar.std_distance()

    '''
    ## Work Done and Fatigue

    Larger increases in Fitness have larger symbols.  Higher fatigue is redder.
    
    Place the cursor over the symbol for more information.
    '''

    df1 = statistics(s, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB, _d(FITNESS_D_ANY))
    if present(df1, _d(FITNESS_D_ANY), pattern=True):
        df1 = coallesce(df1, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB)
        if present(df, TOTAL_CLIMB):
            df1.loc[df1[TOTAL_CLIMB].isna(), [TOTAL_CLIMB]] = 0  # before interpolation
        df2 = statistics(s, FATIGUE_D_ANY, FITNESS_D_ANY)
        df = left_interpolate(df1, df2)
        df[DISTANCE_KM] = df[ACTIVE_DISTANCE] / 1000
        df['Duration'] = df[ACTIVE_TIME].map(format_seconds)
        work_done = sorted_numeric_labels(df.columns, FITNESS)[0]
        fitness = sorted_numeric_labels(df2.columns, FITNESS)[0]
        fatigue = sorted_numeric_labels(df2.columns, FATIGUE)[0]
        print(fatigue, fitness)
        df['FF Ratio'] = df[fatigue] / df[fitness]

        calendar = Calendar(df, title='Work Done and Fatigue', not_hover=[ACTIVE_DISTANCE, ACTIVE_TIME])
        calendar.background('square', fill_alpha=0, line_alpha=1, color='lightgrey')
        calendar.set_palette('FF Ratio', K2R, lo=0.5, hi=2)
        calendar.set_size(work_done, min=0.1, gamma=0.5)
        calendar.foreground('square', fill_alpha=1, line_alpha=0)
        calendar.show()

    '''
    ## Distance, Climb and Direction

    Larger distances have larger symbols.  Higher climbs are redder.  
    The arc indicates the general direction relative to the start.
    
    Place the cursor over the symbol for more information.
    '''

    df = statistics(s, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB, DIRECTION, ASPECT_RATIO)
    df = coallesce(df, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB, DIRECTION, ASPECT_RATIO)
    df[DISTANCE_KM] = df[ACTIVE_DISTANCE] / 1000
    df['Duration'] = df[ACTIVE_TIME].map(format_seconds)
    if present(df, TOTAL_CLIMB):
        df.loc[df[TOTAL_CLIMB].isna(), TOTAL_CLIMB] = 0

    calendar = Calendar(df, title='Distance, Climb and Direction', not_hover=[ACTIVE_DISTANCE, ACTIVE_TIME])
    calendar.std_distance_climb_direction()

    '''
    ## Distance, Work Done and Direction

    Larger distances have larger symbols.  Larger gains in fitness are redder.  
    The arc indicates the general direction relative to the start.
    
    Place the cursor over the symbol for more information.
    '''

    df = statistics(s, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB, DIRECTION, ASPECT_RATIO, _d(FITNESS_D_ANY))
    if present(df, _d(FITNESS_D_ANY), pattern=True):
        df = coallesce(df, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB, DIRECTION, ASPECT_RATIO)
        df[DISTANCE_KM] = df[ACTIVE_DISTANCE] / 1000
        df['Duration'] = df[ACTIVE_TIME].map(format_seconds)
        if present(df, TOTAL_CLIMB):
            df.loc[df[TOTAL_CLIMB].isna(), TOTAL_CLIMB] = 0

        calendar = Calendar(df, title='Distance, Fitness and Direction', not_hover=[ACTIVE_DISTANCE, ACTIVE_TIME])
        calendar.std_distance_fitness_direction()

    '''
    ## Fitness and Fatigue

    Better fitness has larger symbols.  When fatigue is higher symbols have "hotter" colours.
    
    Place the cursor over the symbol for more information.
    '''

    # avoid throwing an exception if missing; plot skipped on next line
    df = statistics(s, FITNESS_D_ANY, FATIGUE_D_ANY, check=False)
    if present(df, FITNESS_D_ANY, pattern=True):
        df = df.resample('1D').mean()
        # take shortest period values when multiple definitions
        fitness = sorted_numeric_labels(df.columns, FITNESS)[0]
        fatigue = sorted_numeric_labels(df.columns, FATIGUE)[0]
        df['FF Ratio'] = df[fatigue] / df[fitness]

        calendar = Calendar(df, title='Fitness and Fatigue', scale=18, border_month=0, border_day=0)
        calendar.set_size(fitness, min=0.1, gamma=0.5)
        calendar.set_palette('FF Ratio', magma(256), lo=0.5, hi=2, min=0)
        calendar.foreground('square', fill_alpha=1, line_alpha=0)
        calendar.show()

    '''
    ## Groups, Distance, Climb and Direction

    Larger distances have larger symbols.  Higher climbs are lighter.  
    The arc indicates the general direction relative to the start.
    Pastel backgrounds group similar rides.
    
    Place the cursor over the symbol for more information.
    '''

    dfa = statistics(s, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB, DIRECTION, ASPECT_RATIO)
    dfa = coallesce(dfa, ACTIVE_DISTANCE, ACTIVE_TIME, TOTAL_CLIMB, DIRECTION, ASPECT_RATIO)
    dfa[DISTANCE_KM] = dfa[ACTIVE_DISTANCE] / 1000
    dfa['Duration'] = dfa[ACTIVE_TIME].map(format_seconds)
    if present(dfa, TOTAL_CLIMB):
        dfa.loc[dfa[TOTAL_CLIMB].isna(), TOTAL_CLIMB] = 0
    dfb = groups_by_time(s)
    if present(dfb, GROUP):
        dfb.loc[dfb[GROUP].isna(), GROUP] = -1
        df = dfa.join(dfb)

        calendar = Calendar(df, not_hover=[ACTIVE_DISTANCE, ACTIVE_TIME], scale=15, border_day=0.1)
        calendar.std_group_distance_climb_direction()




from inspect import getfullargspec
from logging import getLogger
from pkgutil import iter_modules

from .args import SUB_COMMAND, SERVICE, START, STOP, SHOW, JUPYTER, LIST, PROGNAME, NAME, ARG, STATUS
from ..squeal import SystemProcess
from ..uranus import template
from ..uranus.server import JupyterServer, get_controller

log = getLogger(__name__)


def jupyter(args, db):
    '''
## jupyter

    > ch2 jupyter show ...

Show the template in the browser, starting a background Jupyter server if necessary.

    > ch2 jupyter list

List the available templates and their arguments.

    > ch2 jupyter status

Indicate whether the background server is running or not.

    > ch2 jupyter stop

Stop the background server.
    '''
    if args[SUB_COMMAND] == SERVICE:
        service()
    elif args[SUB_COMMAND] == START:
        start()
    elif args[SUB_COMMAND] == STOP:
        stop()
    elif args[SUB_COMMAND] == SHOW:
        show(args)
    elif args[SUB_COMMAND] == STATUS:
        status(db)
    elif args[SUB_COMMAND] == LIST:
        print_list()
    else:
        raise Exception(f'Unexpected command {args[SUB_COMMAND]}')


def service():
    '''
    Start in this thread.
    '''
    get_controller().run_local()


def start():
    get_controller().start_service(restart=True)


def stop():
    get_controller().stop_service()


def status(db):
    with db.session_context() as s:
        if SystemProcess.exists_any(s, JupyterServer):
            print('\n  Service running:')
            url = get_controller().connection_url()
            print(f'    {url}')
            dir = get_controller().notebook_dir()
            print(f'    {dir}\n')
        else:
            print('\n  No service running\n')


def templates():
    log.debug(dir(template))
    log.debug(template.__file__)
    for importer, modname, ispkg in iter_modules(template.__path__):
        module = getattr(template, modname)
        function = getattr(module, modname)
        argspec = getfullargspec(function._original)
        yield modname, (function, argspec)


def print_list():
    for name, (_, spec) in templates():
        args = ' '.join(spec.args)
        print(f'\n  {name}  {args}')
    print()


def show(args):
    name = args[NAME]
    params = args[ARG]
    try:
        fn, spec = dict(templates())[name]
        params = check_params(params, spec)
        fn(*params)
    except KeyError:
        raise Exception(f'No template called {name} (see {PROGNAME} {JUPYTER} {LIST})')


def check_params(params, spec):
    if len(params) != len(spec.args):
        raise Exception(f'Received {len(params)} args but need {len(spec.args)} values ' +
                        f'(see {PROGNAME} {JUPYTER} {LIST})')
    return params





from logging import basicConfig, getLogger, DEBUG
from sys import stdout
from unittest import TestCase


class TestUranus(TestCase):

    def setUp(self):
        if not getLogger().handlers:
            basicConfig(stream=stdout, level=DEBUG)
        self._log = getLogger()

    # def test_template(self):
    #     compare_activities('2018-03-01 16:00', '2017-09-19 16:00', direct=True, group='Bike')

    # def test_display(self):
    #     with TemporaryDirectory() as dir:
    #         try:
    #             self._log.debug(f'Dir {dir}')
    #             set_jupyter_args(self._log, {JUPYTER: True, ROOT: dir})
    #             compare_activities('2018-03-01 16:00', '2017-09-19 16:00', 'Bike', log=self._log)
    #             sleep(5000)
    #         finally:
    #             stop_jupyter(self._log)




from math import log10

from bokeh.io import output_file, show
from bokeh.palettes import Dark2
from bokeh.plotting import figure

from ch2.data import *
from ch2.data.impulse import impulse_10
from ch2.data.response import *
from ch2.squeal import *
from ch2.squeal.database import connect
from ch2.stoats.calculate.impulse import HRImpulse
from ch2.uranus.decorator import template


@template
def fit_ff_parameters(segment_name):

    f'''
    # Fit FF Parameters to {segment_name}

    This notebook allows you to estimate a personal time scale (decay period) for the
    [FF model](https://andrewcooke.github.io/choochoo/impulse) using your times (more exactly, the speed)
    on the named segment.

    The resulting value can be used in the configuration so that the Fitness parameter more accurately
    reflects your personal rate of adaption.
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    s = session('-v2')
    hr10_data = statistics(s, HR_IMPULSE_10, with_sources=True)
    print(hr10_data.describe())
    segment_journals = s.query(SegmentJournal).join(Segment).filter(Segment.name == segment_name).all()
    st_data = statistics(s, SEGMENT_TIME, sources=segment_journals)
    print(st_data.describe())


    '''
    ## Define Initial Model
    
    Define an initial model and resample the data to hours (for faster fitting).
    
    Note that model scale and period are in log units so that the fitter cannot explore negative values.
    
    The default scale is 10 because there are 10 seconds between impulses (I wrote this a long time ago
    and honestly am no longer sure of the logic there).
    
    The default period is 42 days, as used by Training Peaks (note that units are hours because that is
    what we resample the data to).
    '''

    model = DecayModel(1e-10, 0, log10(10), log10(42 * 24), HR_IMPULSE_10, FITNESS)

    '''
    ## Plot Initial Model
    
    Trim and resample data; evaluate and plot our initial model.
    '''

    hr3600_data = pre_calc(hr10_data, model)
    hr3600_data = calc(hr3600_data, model)

    output_file(filename='/dev/null')
    palette = Dark2[3]

    f = figure(plot_width=500, plot_height=450, x_axis_type='datetime')
    f.line(x=hr3600_data.index, y=hr3600_data[FITNESS], color=palette[1], legend=f'new {FITNESS}')
    f.legend.location = 'top_left'
    show(f)

    '''
    ## Fit Model
    
    Trim data and resample data; calculate the target values (speed);
    do an initial evaluation to set the scaling; and then fit.
    '''

    st3600_data = st_data.set_index(st_data.index.round('1H'))
    print(st3600_data.describe())
    hr3600_data = pre_calc(hr10_data, model, target=st3600_data)
    print(hr3600_data.describe())
    data = hr3600_data.join(st3600_data, how='outer')
    data[SPEED] = 1/data[SEGMENT_TIME]

    # for stability we do a progressive adjustment
    model = fit(SPEED, FITNESS, data, model, calc, 'log10_scale', tol=1e-10)
    print(model)
    model = fit(SPEED, FITNESS, data, model, calc, 'log10_scale', 'log10_period', tol=1e-10)
    print(model)
    model = fit(SPEED, FITNESS, data, model, calc, 'log10_scale', 'log10_period', 'zero', tol=1e-10)
    print(model)
    print(f'\nFitted time scale {10**model.log10_period} hr')
    print(f'                  {10**model.log10_period / 24} days')

    '''
    ## Plot Fit
    '''

    results = calc(data, model)

    f = figure(plot_width=500, plot_height=450, x_axis_type='datetime')
    f.line(x=results.index, y=results[FITNESS], color=palette[1], legend=f'new {FITNESS}')
    f.circle(x=st_data.index, y=1.0/st_data[SEGMENT_TIME], fill_color=palette[2], size=10, legend='Speed')
    f.legend.location = 'top_left'
    show(f)

    '''
    ## Stop Here
    
    If running all cells, what follows is not what we want...
    '''

    raise Exception()

    '''
    ## Alternative Impulse Parameters
    
    So, for example, below has gamma=1 instead of the default 2.
    '''

    _, db = connect(['-v5'])
    impulse = HRImpulse(dest_name='Test Impulse', gamma=1.0, zero=2, one=6, max_secs=60)
    hr_data = statistics(s, HR_ZONE)
    print(hr_data.describe())
    #print(hr_data)
    impulse_data = impulse_10(hr_data, impulse)
    print(impulse_data.describe())
    #print(impulse_data)

    '''
    Fit to that data.
    '''

    model = DecayModel(1e-10, 0, log10(10), log10(42 * 24), 'Test Impulse', FITNESS)

    tmp_data = pre_calc(impulse_data, model, target=st3600_data)
    print(tmp_data.describe())
    data = tmp_data.join(st3600_data, how='outer')
    data[SPEED] = 1/data[SEGMENT_TIME]

    model = fit(SPEED, FITNESS, data, model, calc, 'log10_scale', tol=1e-10)
    print(model)
    model = fit(SPEED, FITNESS, data, model, calc, 'log10_scale', 'log10_period', tol=1e-10)
    print(model)
    model = fit(SPEED, FITNESS, data, model, calc, 'log10_scale', 'log10_period', 'zero', tol=1e-10)
    print(model)
    print(f'\nFitted time scale {10**model.log10_period} hr')
    print(f'                  {10**model.log10_period / 24} days')

    '''
    And plot the fit.
    '''

    results = calc(data, model)

    f = figure(plot_width=500, plot_height=450, x_axis_type='datetime')
    f.line(x=results.index, y=results[FITNESS], color=palette[1], legend=f'new {FITNESS}')
    f.circle(x=st_data.index, y=1.0/st_data[SEGMENT_TIME], fill_color=palette[2], size=10, legend='Speed')
    f.legend.location = 'top_left'
    show(f)




from collections import namedtuple
import datetime as dt
from itertools import groupby
from logging import getLogger

import numpy as np

from .frame import linear_resample, present
from ..lib.data import nearest_index, get_index_loc
from ..squeal import StatisticName, StatisticJournal
from ..stoats.names import _d, ELEVATION, DISTANCE, TOTAL_CLIMB, TIME, CLIMB_ELEVATION, CLIMB_DISTANCE, CLIMB_TIME, \
    CLIMB_GRADIENT, CLIMB_POWER, POWER_ESTIMATE, CLIMB_CATEGORY

log = getLogger(__name__)

SCORE = 'Score'
GRADIENT = 'Gradient'

# a climb of 80m is roughly equivalent to a score of 8000 on strava's weird approach -
# https://support.strava.com/hc/en-us/articles/216917057-How-are-Strava-climbs-categorized-For-Rides-
MIN_CLIMB_ELEVATION = 80
MIN_CLIMB_GRADIENT = 3
MAX_CLIMB_GRADIENT = 40
MAX_CLIMB_REVERSAL = 0.1
CLIMB_CATEGORIES = {MIN_CLIMB_ELEVATION: '4', 160: '3', 320: '2', 640: '1', 800: 'HC'}

# trade-off between pure elevation (0) and pure gradient (1)
CLIMB_PHI = 0.6

Climb = namedtuple('Climb', 'phi, min_elevation, min_gradient, max_gradient, max_reversal',
                   defaults=(CLIMB_PHI, MIN_CLIMB_ELEVATION, MIN_CLIMB_GRADIENT,
                             MAX_CLIMB_GRADIENT, MAX_CLIMB_REVERSAL))


def find_climbs(df, params=Climb()):
    df = df.drop_duplicates(subset=[DISTANCE])
    by_dist = df.set_index(df[DISTANCE])
    by_dist = linear_resample(by_dist, quantise=False)
    for dlo, dhi in find_climb_distances(by_dist, params=params):
        tlo, thi = nearest_index(df, DISTANCE, dlo), nearest_index(df, DISTANCE, dhi)
        log.debug(f'Found climb from {tlo} - {thi} ({dlo}m - {dhi}m)')
        up = df[ELEVATION].loc[thi] - df[ELEVATION].loc[tlo]
        along = df[DISTANCE].loc[thi] - df[DISTANCE].loc[tlo]
        climb = {TIME: thi,
                 CLIMB_ELEVATION: up,
                 CLIMB_DISTANCE: along,
                 CLIMB_TIME: (thi - tlo).total_seconds(),
                 CLIMB_GRADIENT: 100 * up / along}
        for height in sorted(CLIMB_CATEGORIES.keys()):
            if up >= height:
                climb[CLIMB_CATEGORY] = CLIMB_CATEGORIES[height]
            else:
                break
        yield climb


def find_climb_distances(df, params=Climb()):
    mn, mx = df[ELEVATION].min(), df[ELEVATION].max()
    if mx - mn > params.min_elevation:
        score, lo, hi = biggest_climb(df, params=params)
        if score:
            a, b, c = split(df, lo, hi)
            yield from find_climb_distances(a, params=params)
            yield from contiguous(b, params=params)
            yield from find_climb_distances(c, params=params)


def split(df, lo, hi, inside=True):
    ilo, ihi = get_index_loc(df, lo), get_index_loc(df, hi)
    if ilo > ihi:
        ilo, ihi = ihi, ilo
    if inside:
        ihi += 1
    else:
        ilo += 1
    return df.iloc[:ilo].copy(), df.iloc[ilo:ihi].copy(), df.iloc[ihi:].copy()


def contiguous(df, params=Climb()):
    up = df[ELEVATION].iloc[-1] - df[ELEVATION].iloc[0]
    if up >= params.min_elevation:
        down, lo, hi = biggest_reversal(df)
        if down and down > params.max_reversal * up:
            a, b, c = split(df, lo, hi, inside=False)
            yield from contiguous(a, params=params)
            yield from find_climb_distances(b, params=params)
            yield from contiguous(c, params=params)
        else:
            along = df.index[-1] - df.index[0]
            if along and 100 * up / along < params.max_gradient:
                yield df.index[0], df.index[-1]


def first_or_none(generator):
    try:
        return next(generator)
    except StopIteration:
        return None


def biggest_reversal(df):
    max_elevation, max_indices, d = 0, (None, None), df.index[1] - df.index[0]
    for offset in range(1, len(df)-1):
        df[_d(ELEVATION)] = df[ELEVATION].diff(-offset)
        if present(df, _d(ELEVATION)):
            d_elevation = df[_d(ELEVATION)].dropna().max()
            if d_elevation > max_elevation:
                max_elevation = d_elevation
                hi = df.loc[df[_d(ELEVATION)] == max_elevation].index[0]  # break ties
                lo = df.index[get_index_loc(df, hi) + offset]
                max_indices = (lo, hi)
    return max_elevation, max_indices[0], max_indices[1]


def biggest_climb(df, params=Climb(), grid=10):
    # returns (score, dlo, dhi))
    # use distances (indices) rather than ilocs because we're subdividing the data
    if len(df) > 100 * grid:
        score, lo, hi = search(df.iloc[::grid].copy())
        if score:
            # need to pass through iloc to extend range
            ilo, ihi = get_index_loc(df, lo), get_index_loc(df, hi)
            return search(df.iloc[max(0, ilo-grid):min(ihi+grid, len(df)-1)].copy(), params=params)
        else:
            return 0, None, None
    else:
        return search(df, params=params)


def search(df, params=Climb()):
    # returns (score, dlo, dhi)
    # use times (indices) rather than ilocs because we're subdividing the data
    max_score, max_indices, d = 0, (None, None), df.index[1] - df.index[0]
    for offset in range(len(df)-1, 0, -1):
        df[_d(ELEVATION)] = df[ELEVATION].diff(offset)
        d_distance = d * offset
        min_elevation = max(params.min_elevation, params.min_gradient * d_distance / 100)
        if df[_d(ELEVATION)].max() > min_elevation:  # avoid some work
            df[SCORE] = df[_d(ELEVATION)] / d_distance ** params.phi
            score = df.loc[df[_d(ELEVATION)] > min_elevation, SCORE].max()
            if not np.isnan(score) and score > max_score:
                max_score = score
                hi = df.loc[df[SCORE] == max_score].index[0]  # arbitrarily pick one if tied (error here w item())
                lo = df.index[get_index_loc(df, hi) - offset]
                max_indices = (lo, hi)
    return max_score, max_indices[0], max_indices[1]


def climbs_for_activity(s, ajournal):

    from ..stoats.calculate.activity import ActivityCalculator

    total = s.query(StatisticJournal).join(StatisticName). \
        filter(StatisticName.name == TOTAL_CLIMB,
               StatisticJournal.time == ajournal.start,
               StatisticName.owner == ActivityCalculator,
               StatisticName.constraint == ajournal.activity_group).order_by(StatisticJournal.time).one_or_none()
    statistics = s.query(StatisticJournal).join(StatisticName). \
        filter(StatisticName.name.like('Climb %'),
               StatisticJournal.time >= ajournal.start,
               StatisticJournal.time <= ajournal.finish,
               StatisticName.owner == ActivityCalculator,
               StatisticName.constraint == ajournal.activity_group).order_by(StatisticJournal.time).all()
    return total, sorted((dict((statistic.statistic_name.name, statistic) for statistic in climb_statistics)
                          for _, climb_statistics in groupby(statistics, key=lambda statistic: statistic.time)),
                         key=lambda climb: climb[CLIMB_ELEVATION].value, reverse=True)


def add_climb_stats(df, climbs):
    for climb in climbs:
        finish = climb[TIME]
        start = finish - dt.timedelta(seconds=climb[CLIMB_TIME])
        climb[CLIMB_POWER] = df.loc[start:finish, [POWER_ESTIMATE]].mean()


# if __name__ == '__main__':
#     from . import *
#     start = time()
#     date = '2017-05-28 10:28:13'  # 1495103293
#     s = session('-v5')
#     df = activity_statistics(s, DISTANCE, ELEVATION, local_time=date, activity_group_name='Bike', with_timespan=False)
#     print(list(find_climbs(df)))
#     print(time() - start)

    # prev
    # Sunday, 28 May 2017 15:26:17
    # Sunday, 28 May 2017 16:03:51
    # Sunday, 28 May 2017 16:47:57
    # Sunday, 28 May 2017 17:19:28
    # Sunday, 28 May 2017 17:38:05

    # this
    # (Timestamp('2017-05-28 15:15:55+0000', tz='UTC'), Timestamp('2017-05-28 15:26:17+0000', tz='UTC'))
    # (Timestamp('2017-05-28 15:31:01+0000', tz='UTC'), Timestamp('2017-05-28 16:03:51+0000', tz='UTC'))
    # (Timestamp('2017-05-28 16:04:23+0000', tz='UTC'), Timestamp('2017-05-28 16:47:57+0000', tz='UTC'))
    # (Timestamp('2017-05-28 17:03:44+0000', tz='UTC'), Timestamp('2017-05-28 17:42:38+0000', tz='UTC'))




import datetime as dt
from logging import getLogger, basicConfig, DEBUG, StreamHandler, Formatter
from sys import stdout
from unittest import TestCase

import sqlalchemy as s
from sqlalchemy.orm import sessionmaker
from urwid import WidgetWrap, Pile, Edit, Filler

from ch2.squeal.binders import Binder
from ch2.squeal.support import Base
from ch2.squeal.types import Date
from ch2.uweird.tui.widgets import Integer

log = getLogger(__name__)

# if not getLogger().handlers:
#     basicConfig(stream=stdout, level=INFO)
# log = getLogger()
# if not log.handlers:
#     log.setLevel(DEBUG)
#     handler = StreamHandler(stdout)
#     handler.setLevel(DEBUG)
#     formatter = Formatter('%(asctime)s - %(levelname)s - %(name)s - %(message)s')
#     handler.setFormatter(formatter)
#     log.addHandler(handler)


class Data(Base):

    __tablename__ = 'data'

    integer = s.Column(s.Integer, primary_key=True)
    text = s.Column(s.Text, nullable=False, server_default='')
    date = s.Column(Date)


class Database:

    def __init__(self):
        self.engine = s.create_engine('sqlite:///:memory:', echo=True)
        self.__create_tables()
        self.session = sessionmaker(bind=self.engine)

    def __create_tables(self):
        log.info('Creating tables')
        Base.metadata.create_all(self.engine)


class DataWidget(WidgetWrap):

    def __init__(self):
        self.integer = Integer('An integer: ')
        self.text = Edit('Some text: ')
        super().__init__(Filler(Pile([self.integer, self.text]), valign='top'))


class TestSqueal(TestCase):

    def test_bind(self):

        db = Database()
        session = db.session()

        data = Data(integer=42, text='abc')
        session.add(data)
        widget = DataWidget()
        binder = Binder(log, session, widget, Data, defaults={'integer': 42})

        self.assertEqual(binder.instance.text, 'abc', binder.instance.text)
        self.assertEqual(widget.text.edit_text, 'abc', widget.text.edit_text)
        self.assertEqual(binder.instance.integer, 42, binder.instance.integer)
        self.assertEqual(widget.integer.state, 42, widget.integer.state)

        # edit the text
        size = (10,)
        for key in ('right', 'right', 'delete'):
            widget.text.keypress(size, key)
        self.assertEqual(widget.text.edit_text, 'ab', widget.text.edit_text)
        self.assertEqual(binder.instance.text, 'ab', binder.instance.text)
        self.assertEqual(binder.instance.integer, 42, binder.instance.integer)

        # modify the primary key
        size = (10, 10)
        for key in ('right', 'right'):
            widget.integer.keypress(size, key)
        with self.assertRaises(Exception):
            widget.integer.keypress(size, 'delete')

        session.expunge_all()
        # no need to add data - it was saved when the first binder above did a query
        widget = DataWidget()
        binder = Binder(log, session, widget, Data, multirow=True, defaults={'integer': 42})

        self.assertEqual(binder.instance.text, 'abc', binder.instance.text)
        self.assertEqual(widget.text.edit_text, 'abc', widget.text.edit_text)
        self.assertEqual(binder.instance.integer, 42, binder.instance.integer)
        self.assertEqual(widget.integer.state, 42, widget.integer.state)

        # edit the text
        size = (10,)
        for key in ('home', 'right', 'right', 'delete'):
            log.info('Keypress %s' % key)
            widget.text.keypress(size, key)
        self.assertEqual(widget.text.edit_text, 'ab', widget.text.edit_text)
        self.assertEqual(binder.instance.text, 'ab', binder.instance.text)
        self.assertEqual(binder.instance.integer, 42, binder.instance.integer)

        # modify the primary key
        size = (10, 10)
        for key in ('right', 'right', 'delete'):
            widget.integer.keypress(size, key)

        self.assertEqual(binder.instance.integer, 4, binder.instance.integer)
        self.assertEqual(binder.instance.text, '', binder.instance.text)

        # and back again
        widget.integer.keypress(size, '2')
        self.assertEqual(binder.instance.integer, 42, binder.instance.integer)
        self.assertEqual(binder.instance.text, 'ab', binder.instance.text)

        # try setting date
        binder.instance.date = dt.date(2018, 7, 1)
        session.commit()
        data = session.query(Data).filter(Data.integer == 42).one()
        self.assertEqual(data.date, dt.date(2018, 7, 1))




from logging import getLogger

from ..lib.date import to_time
from ..squeal import ActivityBookmark
from ..squeal.database import connect

log = getLogger(__name__)


class CoastingBookmark:

    def __init__(self, log, db):
        self._log = log
        self._db = db

    def run(self, min_time, max_cadence, min_speed, constraint=None):
        with self._db.session_context() as s:
            self.__delete_previous(s, constraint)
            for aj_id, start, finish in self.__find(s, min_time, max_cadence, min_speed):
                self.__add(s, aj_id, start, finish, constraint)

    def __delete_previous(self, s, constraint):
        s.query(ActivityBookmark). \
            filter(ActivityBookmark.owner == self,
                   ActivityBookmark.constraint == constraint). \
            delete()

    def __add(self, s, aj_id, start, finish, constraint):
        log.info('%s - %s (%d)' % (to_time(start), to_time(finish), aj_id))
        s.add(ActivityBookmark(activity_journal_id=aj_id, start=start, finish=finish,
                               owner=self, constraint=constraint))

    def __find(self, s, min_time, max_cadence, min_speed):
        s.execute('PRAGMA automatic_index=OFF;')
        for row in s.execute('''
select a.id, s.time, f.time
  from statistic_name as c,
       statistic_name as d,
       statistic_name as v,
       activity_timespan as t cross join
       statistic_journal as s,
       statistic_journal_integer as si,
       statistic_journal as ss,
       statistic_journal_integer as ssi,
       statistic_journal as f,
       statistic_journal_integer as fi,
       statistic_journal as ff,
       statistic_journal_integer as ffi,
       activity_journal as a
 where c.name = 'Cadence'
   and d.name = 'Distance'
   and v.name = 'Speed'
   and s.statistic_name_id = c.id
   and f.statistic_name_id = c.id
   and ss.statistic_name_id = c.id
   and ff.statistic_name_id = c.id
   and s.id = si.id
   and f.id = fi.id
   and ss.id = ssi.id
   and ff.id = ffi.id
   and si.value < :max_cadence
   and fi.value < :max_cadence
   and ss.serial = s.serial-1
   and ff.serial = f.serial+1
   and s.source_id = f.source_id
   and ss.source_id = f.source_id
   and ff.source_id = f.source_id
   and ssi.value >= :max_cadence
   and ffi.value >= :max_cadence
   and f.time - s.time > :min_time
   and t.activity_journal_id = f.source_id
   and t.start < s.time
   and t.finish > f.time
   and a.id == f.source_id
   and not exists (select 1
                     from statistic_journal_integer as ji,
                          statistic_journal as j
                    where ji.value >= :max_cadence
                      and ji.id = j.id
                      and j.statistic_name_id = c.id
                      and j.source_id = s.source_id
                      and j.serial > s.serial
                      and j.serial < f.serial)
   and exists (select 1
                 from statistic_journal_float as f1,
                      statistic_journal_float as f2,
                      statistic_journal as j1,
                      statistic_journal as j2
                where f1.id = j1.id
                  and f2.id = j2.id
                  and j1.statistic_name_id = d.id
                  and j2.statistic_name_id = d.id
                  and j1.serial = s.serial
                  and j2.serial = f.serial
                  and j1.source_id = f.source_id
                  and j2.source_id = f.source_id
                  and f2.value - f1.value > :min_speed * (f.time - s.time))
   and not exists (select 1
                     from statistic_journal_float as jf,
                          statistic_journal as j
                    where jf.value = 0
                      and jf.id = j.id
                      and j.statistic_name_id = v.id
                      and j.source_id = s.source_id
                      and j.serial >= s.serial
                      and j.serial <= f.serial);
        ''', {'min_time': min_time, 'max_cadence': max_cadence, 'min_speed': min_speed}):
            yield row[0], row[1], row[2]
        s.execute('PRAGMA automatic_index=ON;')


if __name__ == '__main__':
    '''
    Bookmark sections of activities that exceed 60s of little or no pedalling (maximum cadence 20)
    and with any speed (max 0).
    
    These are then used in the fit_power_parameters template to estimate CdA and Crr. 
    '''
    ns, db = connect(['-v 4'])
    CoastingBookmark(log, db).run(60, 20, 0, constraint='60/20/0')
#     CoastingBookmark(log, db).run(15, 5, 10, constraint='15/5/10')




from logging import getLogger, NullHandler
from sys import version_info

getLogger('bokeh').addHandler(NullHandler())
getLogger('tornado').addHandler(NullHandler())


class FatalException(Exception):

    '''
    Base class for exceptions that we can't ignore at some higher level
    (fundamental things like bad config).
    '''

    pass


from .commands.activities import activities
from .commands.args import COMMAND, parser, NamespaceWithVariables, PROGNAME, HELP, DEV, DIARY, FIT, \
    PACKAGE_FIT_PROFILE, ACTIVITIES, NO_OP, CONFIG, CONSTANTS, STATISTICS, TEST_SCHEDULE, MONITOR, GARMIN, \
    UNLOCK, DUMP, FIX_FIT, CH2_VERSION, JUPYTER
from .commands.constants import constants
from .commands.dump import dump
from .commands.config import config
from .commands.diary import diary
from .commands.fit import fit
from .commands.fix_fit import fix_fit
from .commands.garmin import garmin
from .commands.jupyter import jupyter
from .commands.help import help, LengthFmt
from .commands.monitor import monitor
from .commands.package_fit_profile import package_fit_profile
from .commands.statistics import statistics
from .commands.test_schedule import test_schedule
from .commands.unlock import unlock
from .lib.io import tui
from .lib.log import make_log, log_current_exception
from .squeal.database import Database
from .squeal import SystemConstant
from .uranus.server import start_controller

log = getLogger(__name__)


@tui
def no_op(args, db):
    '''
## no-op

This is used internally when accessing data in Jupyter or configuring the system
at the command line.
    '''
    pass


COMMANDS = {ACTIVITIES: activities,
            CONSTANTS: constants,
            CONFIG: config,
            DIARY: diary,
            DUMP: dump,
            FIT: fit,
            FIX_FIT: fix_fit,
            GARMIN: garmin,
            JUPYTER: jupyter,
            HELP: help,
            MONITOR: monitor,
            STATISTICS: statistics,
            NO_OP: no_op,
            PACKAGE_FIT_PROFILE: package_fit_profile,
            TEST_SCHEDULE: test_schedule,
            UNLOCK: unlock,
            }


def main():
    args = NamespaceWithVariables(parser().parse_args())
    command_name = args[COMMAND] if COMMAND in args else None
    command = COMMANDS[command_name] if command_name in COMMANDS else None
    tui = command and hasattr(command, 'tui') and command.tui
    make_log(args, tui=tui)
    log.info('Version %s' % CH2_VERSION)
    if version_info < (3, 7):
        raise Exception('Please user Python 3.7 or more recent')
    db = Database(args, log)
    try:
        if db.is_empty() and (not command or command_name not in (CONFIG, PACKAGE_FIT_PROFILE)):
            refuse_until_configured()
        elif command:
            start_controller(db, args)
            command(args, db)
        else:
            log.debug('If you are seeing the "No command given" error during development ' +
                      'you may have forgotten to set the command name via `set_defaults()`.')
            raise Exception('No command given (try `ch2 help`)')
    except KeyboardInterrupt:
        log.critical('User abort')
        exit(1)
    except Exception as e:
        log.critical(e)
        log_current_exception()
        log.info('See `%s %s` for available commands.' % (PROGNAME, HELP))
        log.info('Docs at http://andrewcooke.github.io/choochoo')
        if not args or args[DEV]:
            raise
        exit(2)


def refuse_until_configured():
    LengthFmt().print_all('''
Welcome to Choochoo.

Before using the ch2 command you must configure the system.

Please see the documentation at http://andrewcooke.github.io/choochoo

To generate a default configuration use the command

    %s %s

NOTE: The default configuration is only an example.  Please see the docs
for more details.''' % (PROGNAME, CONFIG))




from collections import namedtuple
from logging import getLogger
from math import pi

# hide from imports of this package
import numpy as _np
import pandas as _pd
import scipy as _sp

from .frame import median_dt, session, activity_statistics, bookmarks
from .lib import fit, inplace_decay
from ..lib.data import tmp_name
from ..stoats.names import *
from ..stoats.names import _d, _sqr, _avg

log = getLogger(__name__)
RAD_TO_DEG = 180 / pi


def add_differentials(df, max_gap=None):
    return _add_differentials(df, SPEED, DISTANCE, ELEVATION, SPEED, SPEED_2, LATITUDE, LONGITUDE,
                              max_gap=max_gap)


def add_air_speed(df, wind_speed=0, wind_heading=0, max_gap=None):
    df[AIR_SPEED] = df[SPEED] + wind_speed * _np.cos((df[HEADING] - wind_heading) / RAD_TO_DEG)
    return _add_differentials(df, AIR_SPEED, AIR_SPEED, max_gap=max_gap)


def _add_differentials(df, speed, *names, max_gap=None):

    # rather than use timespans (old approach) it's more reliable to discard any intervals
    # over a certain time gap.

    speed_2 = _sqr(speed)
    df[speed_2] = df[speed] ** 2

    tmp = tmp_name()
    df[tmp] = df.index
    df[tmp] = df[tmp].diff().dt.seconds

    for name in names:
        df[_d(name)] = df[name].diff()
        df.loc[df[tmp] > max_gap, [_d(name)]] = _np.nan

    if LATITUDE in names and LONGITUDE in names and HEADING not in df.columns:
        df[HEADING] = _np.arctan2(df[_d(LONGITUDE)], df[_d(LATITUDE)]) * RAD_TO_DEG
        df.loc[df[tmp] > max_gap, [HEADING]] = _np.nan

    avg_speed_2 = [(a**2 + a*b + b**2)/3 for a, b in zip(df[speed], df[speed][1:])]
    df[_avg(speed_2)] = [_np.nan] + avg_speed_2
    if max_gap:
        df.loc[df[tmp] > max_gap, [_avg(speed_2)]] = _np.nan

    df.drop(columns=[tmp], inplace=True)
    return df


def _add_differentials_old(df, speed, *names):

    # todo - remove once we've checked power calcs

    speed_2 = _sqr(speed)
    df[speed_2] = df[speed] ** 2

    def diff():
        for _, old_span in df.groupby(TIMESPAN_ID):
            # discard leading and trailing na
            subset = old_span[list(names)].isna().any(axis=1).replace(True, _np.nan)
            start, finish = subset.first_valid_index(), subset.last_valid_index()
            if start and finish:
                old_span = old_span.loc[start:finish]

                if all(len(old_span[name]) == len(old_span[name].dropna()) for name in names):
                    new_span = _pd.DataFrame(index=old_span.index)
                    for col in names:
                        new_span[_d(col)] = old_span[col].diff()
                    if HEADING not in old_span.columns:
                        new_span[HEADING] = _np.arctan2(new_span[_d(LONGITUDE)], new_span[_d(LATITUDE)]) * RAD_TO_DEG
                    avg_speed_2 = [(a**2 + a*b + b**2)/3 for a, b in zip(old_span[speed], old_span[speed][1:])]
                    new_span[_avg(speed_2)] = [_np.nan] + avg_speed_2
                    yield new_span

    spans = list(diff())
    if len(spans):
        extra = _pd.concat(spans).sort_index()
        return df.drop(columns=list(extra.columns), errors='ignore').join(extra)
    else:
        raise PowerException('Missing data - found no spans without NANs')


def add_energy_budget(df, m, g=9.8):
    # if DELTA_ELEVATION is +ve we've gone uphill.  so this is the total amount of energy
    # gained in this segment.
    df[DELTA_ENERGY] = m * (df[DELTA_SPEED_2] / 2 + df[DELTA_ELEVATION] * g)
    return df


def add_cda_estimate(df, p=1.225):
    # https://www.cyclingpowerlab.com/CyclingAerodynamics.aspx
    # assume that all energy lost (-ve gain) is due to air resistance.
    df[CDA] = -df[DELTA_ENERGY] / (p * df[AVG_AIR_SPEED_2] * df[DELTA_DISTANCE] * 0.5)
    return df


def add_crr_estimate(df, m, g=9.8):
    # assume that all energy lost is due to rolling resistance
    df[CRR] = -df[DELTA_ENERGY] / (df[DELTA_DISTANCE] * m * g)
    return df


def add_loss_estimate(df, m, cda=0.45, crr=0, p=1.225, g=9.8):
    # this is the energy spent on air and rolling resistance
    df[LOSS] = (cda * p * df[AVG_AIR_SPEED_2] * 0.5 + crr * m * g) * df[DELTA_DISTANCE]
    return df


def add_power_estimate(df):
    # power input must balance the energy budget.
    df[POWER_ESTIMATE] = (df[DELTA_ENERGY] + df[LOSS]) / df[DELTA_TIME].dt.total_seconds()
    df[POWER_ESTIMATE].clip(lower=0, inplace=True)
    if CADENCE in df.columns: df.loc[df[CADENCE] < 1, [POWER_ESTIMATE]] = 0
    df.loc[df[POWER_ESTIMATE].isna(), [POWER_ESTIMATE]] = 0
    energy = (df[POWER_ESTIMATE].iloc[1:] * df[DELTA_TIME].iloc[1:]).cumsum()
    df[ENERGY] = 0
    df.loc[1:, [ENERGY]] = energy
    return df


def add_modeled_hr(df, window, slope, delay):
    dt = median_dt(df)
    window, delay = int(0.5 + window / dt), delay / dt
    df[DETRENDED_HEART_RATE] = df[HEART_RATE] - df[HEART_RATE].rolling(window, center=True, min_periods=1).median()
    df[PREDICTED_HEART_RATE] = df[POWER_ESTIMATE] * slope
    inplace_decay(df, PREDICTED_HEART_RATE, delay)
    df[PREDICTED_HEART_RATE] -= df[PREDICTED_HEART_RATE].rolling(window, center=True, min_periods=1).median()
    return df


def measure_initial_delay(df, dt=None, col1=HEART_RATE, col2=POWER_ESTIMATE, n=20):
    dt = dt or median_dt(df)
    correln = [(i, df[col1].corr(df[col2].shift(freq=f'{i * dt}S'))) for i in range(-n, n + 1)]
    correln = sorted(correln, key=lambda c: c[1], reverse=True)
    return dt * correln[0][0]


def measure_initial_scaling(df):
    delay = measure_initial_delay(df)
    if delay < 0: raise PowerException('Cannot estimate delay (insufficient data?)')
    df[DELAYED_POWER] = df[POWER_ESTIMATE].shift(freq=f'{delay}S')
    clean = df.loc[:, (DELAYED_POWER, HEART_RATE)].dropna()
    fit = _sp.stats.linregress(x=clean[DELAYED_POWER], y=clean[HEART_RATE])
    log.debug(f'Initial fit {fit}')
    return fit.slope, fit.intercept,  delay


class PowerException(Exception): pass


# the 13.5 delay is from fitting my rides til 2019
# it seems oddly low to me - i wonder if i have an error confusing bins and time
PowerModel = namedtuple('PowerModel', 'cda, crr, slope, window, delay, m,  wind_speed, wind_heading',
                        defaults=[0.5, 0,   200,   60*60,  13.5,  70, 0,          0])


def evaluate(df, model, quiet=True):
    if not quiet: log.debug(f'Evaluating {model}')
    df = add_energy_budget(df, model.m)
    df = add_air_speed(df, model.wind_speed, model.wind_heading)
    df = add_loss_estimate(df, model.m, cda=model.cda, crr=model.crr)
    df = add_power_estimate(df)
    return df


MIN_DELAY = 1


def fit_power(df, model, *vary, tol=0.1):

    log.debug(f'Fit power: varying {vary}')
    df = evaluate(df, model)
    slope, intercept, delay = measure_initial_scaling(df)
    # we ignore intercept because removing the median makes it very weakly constrained
    model = model._replace(slope=slope, delay=delay)
    log.debug(f'Fit power: initial model {model}')

    # the internal delay is continuous
    # model delay is MIN_DELAY when the internal delay is zero, and otherwise increases

    def forwards(kargs):
        if 'delay' in kargs:
            kargs['delay'] = kargs['delay'] - MIN_DELAY
        return kargs

    def backwards(kargs):
        if 'delay' in kargs:
            kargs['delay'] = abs(kargs['delay']) + MIN_DELAY
        return kargs

    def evaluate_and_extend(df, model):
        df = evaluate(df, model)
        df = add_modeled_hr(df, model.window, model.slope, model.delay)
        return df

    model = fit(DETRENDED_HEART_RATE, PREDICTED_HEART_RATE, df, model, evaluate_and_extend,
                *vary, forwards=forwards, backwards=backwards, tol=tol)

    log.debug(f'Fit power: model before fixing {model}')
    if model.wind_speed < 0:
        model = model._replace(wind_speed=abs(model.wind_speed), wind_heading=model.wind_heading+180)
    model = model._replace(wind_heading=model.wind_heading % 360)

    log.debug(f'Fit power: final model {model}')
    return model


if __name__ == '__main__':
    s = session('-v 5')
    route = activity_statistics(s, LATITUDE, LONGITUDE, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DISTANCE,
                                ELEVATION, SPEED, CADENCE, bookmarks=bookmarks(s, '60/20/0'), with_timespan=True)
    route.sort_index(inplace=True)
    route = add_differentials(route)  # todo max_gap
    print(route.describe())
    print(route.loc[route[DELTA_DISTANCE] > 1000])



from collections import OrderedDict
from collections.abc import Sequence

from urwid import emit_signal, connect_signal, Widget, ExitMainLoop, disconnect_signal

from .focus import Focus, FocusAttr, FocusWrap
from ...lib.utils import force_iterable


# new tab manager design

# tabs can be arranged in groups.  group contents can be wiped and re-added.
# this allows tabs in the "middle" of a traverse to be rebuilt.
# tabs are assembled in a TabList and then passed to a TabNode which
# contains the group.  the TabNode has to be a widget itself since it
# needs to re-raise signals for tabbing.  because it is a WidgetWrap we need
# the intermediate TabList to assemble the group contents (since the
# TabNode will often be created later).  groups can be nested (a TabNode
# can appear in a TabList) and will behave correctly.  the top-most TabNode
# must have discover() called to set signals for tab looping and to discover
# focuses.

# the functionality depends on Focus.apply taking a keypress argument which
# is duplicated by TabNodes.  on TabNodes this triggers internal logic.

# the widget tree traversal depends on code following conventions that
# i am inferring from the urwid code, but which aren't explicit anywhere.
# in particular, ever widget subclass should either:
# - be a decorator, in which case base_widget is used
# - or be a collection, exposing contents and focus_position
# the only exceptions are "leaf" widgets - tabs wil not move inside these.

# to get this functionality with wrapped widgets, use FocusWrap.


class Tab(FocusWrap):
    """
    A widget wrapper that is added automatically by TagList.add().  Must
    be added to any node that is both target and source of tabbing.
    Intercepts tab keypresses and raises a signal that causes the focus to
    change.

    Normal use is:
        tabs = TabList()
        ...
        widget = tabs.add(Widget(...))
    """

    def __init__(self, w):
        super().__init__(w)

    signals = ['tab']

    def keypress(self, size, key):
        # todo - pass to super first and only handle tabs that are not handled
        # by the widget?
        if key in ('tab', 'shift tab'):
            emit_signal(self, 'tab', self, key)
        else:
            return super().keypress(size, key)


class TabList(Sequence):
    """
    A list of tabbed widgets (in tabbing order) that will be managed by a TabNode.
    The list allows these to be assembled before the TabNode instance is created.
    May include both widgets and other TabNode instances.
    """

    def __init__(self):
        """
        Create an empty list.
        """
        self.__tabs = []

    def __wrap(self, widget_or_node):
        ignore = isinstance(widget_or_node, TabNode) or isinstance(widget_or_node, Tab)
        return widget_or_node if ignore else Tab(FocusAttr(widget_or_node))

    def append(self, widget_or_node):
        """
        Add a widget to the list of managed widgets.  The return value should be
        used in the constructed tree of widgets (it contains both a Tab target and
        a FocusAttr).
        """
        # todo - how do we modify FocusAttr?
        widget_or_node = self.__wrap(widget_or_node)
        self.__tabs.append(widget_or_node)
        return widget_or_node

    def remove(self, tab):
        self.__tabs.remove(tab)

    def insert(self, index, widget_or_node):
        widget_or_node = self.__wrap(widget_or_node)
        self.__tabs.insert(index, widget_or_node)
        return widget_or_node

    def insert_many(self, index, tab_list):
        for tab in reversed(tab_list):
            self.__tabs.insert(index, self.__wrap(tab))
        return tab_list

    def __getitem__(self, item):
        return self.__tabs[item]

    def __setitem__(self, index, item):
        self.__tabs[index] = item

    def __len__(self):
        return len(self.__tabs)


class TabNode(FocusWrap):
    """
    A widget wrapper that encapsulates a (local) root node in the widget tree and
    manages all the tabs below that node.

    In dynamic applications the entire TabList may be replaced using replace_all().
    If only a subset of all nodes need to be replaced, use a nested TabNode (so
    the entire contents of the nested node are replaced).

    Normal use is:

        tabs = TabList()
        widget1 = tabs.append(Widget(...))
        ...
        widgetN = tabs.append(Widget(...))
        root = TabNode(Container([widget1, ... windgetN]), tabs)
        root.discover()
    """

    signals = ['tab']

    def __init__(self, log, widget, tab_list):
        """
        Create a (local) root to the widget tree that manages tabs to the widgets
        below (possibly via nested TabNode instances).
        """
        super().__init__(widget)
        self.__log = log
        self.__tabs_and_indices = {}
        self.__focus = OrderedDict()
        self.__root = None
        self.__path = None
        self.__build_data(tab_list)

    def __build_data(self, tab_list):
        for tab in tab_list:
            n = len(self)
            self.__tabs_and_indices[tab] = n
            self.__tabs_and_indices[n] = tab
            self.__focus[tab] = None
            disconnect_signal(tab, 'tab', self.tab)
            connect_signal(tab, 'tab', self.tab)

    def __to_tab_list(self):
        tab_list = TabList()
        for i in range(len(self)):
            tab_list.append(self.__tabs_and_indices[i])
        return tab_list

    def replace(self, tabs):
        """
        Replace all the managed tabs.  Typically used at the local root of a dynamic
        section of the widget tree.
        """
        self.__tabs_and_indices = {}
        self.__focus = OrderedDict()
        self.__build_data(tabs)
        self.discover(root=self.__root, path=self.__path)

    def insert(self, index, tab):
        """
        Add a tab at the index give (so insert(0, tab) is at start)
        """
        tab_list = self.__to_tab_list()
        tab_list.insert(index, tab)
        self.replace(tab_list)

    def insert_many(self, index, tabs):
        tab_list = self.__to_tab_list()
        tab_list.insert_many(index, tabs)
        self.replace(tab_list)

    def __len__(self):
        return len(self.__focus)

    def tab(self, tab, key):
        """
        The target for tab signals from managed Tab() instances.

        On receiving a signal:
        * check whether tabbing can be handled locally and, if so, activate
        * check if we are root and, if so, loop around
        * otherwise re-raise to tab to remote neighbours (from nested node)
        """
        delta = 1 if key == 'tab' else -1
        n = self.__tabs_and_indices[tab] + delta
        if 0 <= n < len(self):
            self.__try_set_focus(n, key)
        elif not self.__path:
            self.to(None, key)  # loop around
        else:
            emit_signal(self, 'tab', self, key)

    def __try_set_focus(self, n, key):
        try:
            self.__log.debug('Trying to set focus on %s' % self.__tabs_and_indices[n])
            self.__set_focus(self.__tabs_and_indices[n], key)
        except AttributeError:
            self.discover(root=self.__root, path=self.__path)
            self.__set_focus(self.__tabs_and_indices[n], key)

    def __set_focus(self, tab, key):
        self.__log.debug('Using %s' % self.__focus[tab])
        self.__focus[tab].to(self.__root, key)

    def to(self, unused, key):
        """
        Replicate the Focus() interface.  This is used internally for sub-nodes.
        Instead of assigning focus using Focus.to(),
        """
        if self.__focus:
            n = 0 if key == 'tab' else len(self) - 1
            self.__log.debug('Re-targetting at %d' % n)
            self.__try_set_focus(n, key)
        else:
            # we have nothing to focus, so re-raise signal for remote neighbours
            self.__log.debug('Empty so raise signal')
            emit_signal(self, 'tab', self, key)

    def discover(self, root=None, path=None, discard=False):
        """
        Register the root widget here before use (in many cases the root node is
        also this TabNode, so no root argument is needed).

        Does a search of the entire widget tree, recording paths to added widgets
        so that they can be given focus quickly.
        """
        self.__root = root if root else self
        self.__path = path if path else []
        stack = [(self, self.__path)]  # only search sub-tree, not from root

        def unpack_container(widget, path):
            # contents can be dict, list or ListBox
            try:
                iterator = widget.contents.items()
            except AttributeError:
                try:
                    iterator = enumerate(widget.contents)
                except TypeError:
                    iterator = widget.contents.body.items()
            for (key, data) in iterator:
                # data can be widget or tuple containing widget
                yield force_iterable(data), list(path) + [key]

        def unpack_decorator(widget):
            if hasattr(widget, '_wrapped_widget'):
                self.__log.warning('Widget %s (type %s) doesn\'t expose contents' % (widget, type(widget)))
            elif hasattr(widget, 'base_widget'):
                if widget == widget.base_widget:
                    self.__log.debug('Widget with no focus: %s (type %s)' % (widget, type(widget)))
                else:
                    yield widget.base_widget

        def handle_new_widget(widget, path):
            if widget in self.__focus:
                if isinstance(widget, TabNode):
                    self.__focus[widget] = widget
                    widget.discover(self.__root, path=path)
                else:
                    self.__focus[widget] = Focus(path, self.__log)
            else:
                stack.append((widget, path))

        while stack:
            widget, path = stack.pop()
            try:
                for data, new_path in unpack_container(widget, path):
                    for datum in data:
                        if isinstance(datum, Widget):  # filter junk in data
                            handle_new_widget(datum, new_path)
            except AttributeError:
                for widget in unpack_decorator(widget):
                    handle_new_widget(widget, path)

        missing = list(filter(lambda x: self.__focus[x] is None, self.__focus))
        if missing:
            if discard:
                self.__log.info('Removing %s tabs' % len(missing))
                self.remove_many(missing)
            else:
                msg = map(lambda w: '%s (base type %s)' % (w, type(w._w.base_widget)), missing)
                raise Exception('Could not find %s' % ', '.join(msg))


class Root(TabNode):

    def __init__(self, log, widget, tab_list, quit='meta q', save='meta s', abort='meta x', session=None):
        super().__init__(log, widget, tab_list)
        self.__quit = quit
        self.__save = save
        self.__abort = abort
        self.__session = session

    def keypress(self, size, key):
        if key == self.__quit:
            self.save()
            raise ExitMainLoop()
        elif key == self.__abort:
            raise ExitMainLoop()
        elif key == self.__save:
            self.save()
        else:
            return super().keypress(size, key)

    def save(self):
        if self.__session:
            self.__session.flush()
            self.__session.commit()




from .data import sorted_numeric_labels, left_interpolate, bookend
from .date import local_date_to_time, to_date, time_to_local_time, format_date, format_seconds, format_time, \
    local_time_to_time, time_to_local_date, to_duration, to_time, datetime_to_epoch
from .utils import short_str, sign, format_metres, format_percent, format_watts




from logging import getLogger

from sqlalchemy import not_

from . import ActivityJournalCalculatorMixin, MultiProcCalculator, DataFrameCalculatorMixin
from ..names import SEGMENT_TIME, S, summaries, MIN, MSR, CNT, HEART_RATE, \
    SEGMENT_HEART_RATE, BPM, MAX
from ...data import activity_statistics, present, linear_resample_time
from ...squeal import ActivityJournal, SegmentJournal, StatisticJournalFloat, Timestamp

log = getLogger(__name__)

SJOURNAL = 'sjournal'


class SegmentCalculator(ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def _missing(self, s):
        # extends superclass with restriction on activities that have a segment
        existing_ids = s.query(Timestamp.key). \
            filter(Timestamp.owner == self.owner_out)
        activity_ids = s.query(SegmentJournal.activity_journal_id)
        q = s.query(ActivityJournal.start). \
            filter(not_(ActivityJournal.id.in_(existing_ids)),
                   ActivityJournal.id.in_(activity_ids))
        return [row[0] for row in self._delimit_query(q)]

    def _read_dataframe(self, s, ajournal):
        return activity_statistics(s, HEART_RATE, activity_journal=ajournal)

    def _calculate_stats(self, s, ajournal, df):
        all = []
        for sjournal in s.query(SegmentJournal). \
                filter(SegmentJournal.activity_journal == ajournal).all():
            stats = {SJOURNAL: sjournal,
                     SEGMENT_TIME: (sjournal.finish - sjournal.start).total_seconds()}
            if present(df, HEART_RATE):
                ldf = linear_resample_time(df, start=sjournal.start, finish=sjournal.finish)
                stats[SEGMENT_HEART_RATE] = ldf[HEART_RATE].mean()
            all.append(stats)
        return all

    def _copy_results(self, s, ajournal, loader, all):
        for stats in all:
            sjournal = stats[SJOURNAL]
            loader.add(SEGMENT_TIME, S, summaries(MIN, CNT, MSR), sjournal.segment, sjournal,
                       stats[SEGMENT_TIME], sjournal.start, StatisticJournalFloat)
            if SEGMENT_HEART_RATE in stats:
                loader.add(SEGMENT_HEART_RATE, BPM, summaries(MAX, CNT, MSR), sjournal.segment, sjournal,
                           stats[SEGMENT_HEART_RATE], sjournal.start, StatisticJournalFloat)




from logging import getLogger
from unittest import TestCase

from ch2.data import session
from ch2.data.constraint import constraint, build_activity_query

log = getLogger(__name__)


class TestConstraints(TestCase):

    def test_example(self):
        # test requires existing database with stats
        s = session('-v5')
        ast = constraint('Active Distance > 10000')[0]
        q = build_activity_query(s, ast)
        print(q)
        print(list(q))





import datetime as dt
from unittest import TestCase

from ch2.fit.format.read import filtered_records
from ch2.fit.profile.profile import read_fit
from ch2.lib.date import to_time, format_time, local_time_to_time, time_to_local_time


class TestDate(TestCase):

    def test_to_datetime(self):
        self.assertEqual(format_time(to_time('2810-09-21 13:24:01')), '2810-09-21 13:24:01')
        self.assertEqual(format_time(to_time('2810-09-21 13:24')), '2810-09-21 13:24:00')
        self.assertEqual(format_time(to_time('2810-09-21')), '2810-09-21 00:00:00')

    def test_tz(self):
        time = local_time_to_time('2019-02-16 12:19:00')
        self.assertEqual(time, to_time('2019-02-16 15:19:00'))
        self.assertEqual(time_to_local_time(time), '2019-02-16 12:19:00')

    def test_date16(self):
        data = read_fit('data/test/source/other/38014592427.fit')
        # types, messages, records = filtered_records(data, after_records=2342)
        types, messages, records = filtered_records(data)
        previous = None
        for (a, b, record) in records:
            if 'timestamp' in record._fields:
                if previous and previous > record.timestamp:
                    # we have time-travel of 1 minute in this data
                    self.assertTrue(previous - record.timestamp < dt.timedelta(minutes=2),
                                    f'{previous} -> {record.timestamp}')
                previous = record.timestamp




from textblob import TextBlob


def wordlist(source, src, dest=None):
    if not dest:
        dest = src
    source[dest] = source[src].apply(lambda text: TextBlob(text).words if text is not None else [])
    source[dest] = source[dest].apply(lambda words: [w.lower() for w in words])


class WeightedScorer:

    def __init__(self, intensifiers, terms, gamma=0):
        self._intensifiers = intensifiers
        self._terms = terms
        self._gamma = gamma

    def score_words(self, words):
        score = 0
        for i, word in enumerate(words):
            delta = 0
            if word in self._terms:
                if i and words[i-1] in self._intensifiers:
                    if not i-1 or words[i-2] != 'no':
                        delta = self._intensifiers[words[i-1]] * self._terms[word]
                elif not i or words[i-1] != 'no':
                    delta = self._terms[word]
                # print(word, delta)
                score += delta
        if words:
            return score / len(words) ** self._gamma
        else:
            return 0

    def score_all(self, source, words, score):
        source[score] = source[words].apply(self.score_words)






from unittest import TestCase

from ch2.data.constraint import constraint
from ch2.lib.peg import literal, drop, sequence, choice, pattern


class TestPeg(TestCase):

    def test_literal(self):
        self.assertEqual(list(literal('ab')('abc')), [(['ab'], 'c')])

    def test_drop(self):
        self.assertEqual(list(drop(literal('ab'))('abc')), [([], 'c')])

    def test_sequence(self):
        a = literal('a')
        b = literal('b')
        self.assertEqual(list(sequence(a, b)('abc')), [(['a', 'b'], 'c')])

    def test_choice(self):
        a = literal('a')
        b = literal('b')
        self.assertEqual(list(choice(a, b)('abc')), [(['a'], 'bc')])

    def test_pattern(self):
        self.assertEqual(list(pattern(r'(\d+)')('123x')), [(['123'], 'x')])
        self.assertEqual(list(pattern(r'\d+')('123x')), [([], 'x')])
        self.assertEqual(list(pattern(r'\d(\d+)')('123x')), [(['23'], 'x')])

    def test_name(self):
        self.assertEqual(list(constraint('Active Distance > 10')), [('Active Distance', '>', 10.0)])

    def test_term(self):
        self.assertEqual(list(constraint('a = "b" & (c <= 2 | 1.2 > e)')),
                         [(('a', '=', 'b'), '&', (('c', '<=', 2.0), '|', ('e', '<', 1.2)))])
        self.assertEqual(list(constraint('a = "b" & c <= 2 | 1.2 > e')),
                         [((('a', '=', 'b'), '&', ('c', '<=', 2.0)), '|', ('e', '<', 1.2))])




from logging import getLogger

import numpy as np
import pandas as pd
from math import atan2, cos, sin, sqrt, pi

from .frame import linear_resample, median_dt, present, linear_resample_time
from ..stoats.names import HEART_RATE, MAX_MED_HR_M, POWER_ESTIMATE, ACTIVE_DISTANCE, ACTIVE_TIME, \
    ACTIVE_SPEED, TIMESPAN_ID, TIME, DISTANCE, MIN_KM_TIME, MED_KM_TIME, PERCENT_IN_Z, TIME_IN_Z, HR_ZONE, \
    MAX_MEAN_PE_M, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, DIRECTION, ASPECT_RATIO

log = getLogger(__name__)

MAX_MINUTES = (5, 10, 30, 60, 90, 120, 180)


def round_km():
    yield from range(5, 21, 5)
    yield from range(25, 76, 25)
    yield from range(100, 251, 50)
    yield from range(300, 1001, 100)


def active_stats(df):
    stats = {ACTIVE_DISTANCE: 0, ACTIVE_TIME: 0, ACTIVE_SPEED: 0}
    for timespan in df[TIMESPAN_ID].dropna().unique():
        slice = df.loc[df[TIMESPAN_ID] == timespan]
        stats[ACTIVE_DISTANCE] += slice[DISTANCE].max() - slice[DISTANCE].min()
        stats[ACTIVE_TIME] += (slice.index.max() - slice.index.min()).total_seconds()
    stats[ACTIVE_SPEED] = 3.6 * stats[ACTIVE_DISTANCE] / stats[ACTIVE_TIME]
    return stats


def times_for_distance(df, km=None, delta=10):
    stats, km = {}, km or round_km()
    try:
        tmp = pd.DataFrame({TIME: df.index}, index=df[DISTANCE])
        tmp = tmp[~tmp.index.duplicated(keep='last')]
        t4d = pd.DataFrame({TIME: (tmp[TIME] - tmp[TIME].iloc[0]).astype(np.int64) / 1e9}, index=df[DISTANCE])
        lt4d = linear_resample(t4d, d=delta)
        for target in km:
            n = target * 1000 / delta
            dlt4d = lt4d.diff(periods=n).dropna()
            if present(dlt4d, TIME):
                stats[MIN_KM_TIME % target] = dlt4d[TIME].min()
                stats[MED_KM_TIME % target] = dlt4d[TIME].median()
    except Exception as e:
        log.warning(f'No Time for Distance stats: {e}')
    return stats


def hrz_stats(df, zones=None):
    stats, zones = {}, zones or range(1, 8)
    try:
        if present(df, HR_ZONE):
            ldf = linear_resample_time(df, with_timespan=True)
            hrz = pd.cut(ldf[HR_ZONE], bins=zones, right=False).value_counts()
            dt, total = median_dt(ldf), hrz.sum()
            for interval, count in hrz.iteritems():
                zone = interval.left
                stats[PERCENT_IN_Z % zone] = 100 * count / total
                stats[TIME_IN_Z % zone] = dt * count
    except Exception as e:
        log.warning(f'No HR stats: {e}')
    return stats


def max_mean_stats(df, params=((POWER_ESTIMATE, MAX_MEAN_PE_M),), mins=None, delta=10, zero=0):
    stats, mins = {}, mins or MAX_MINUTES
    try:
        ldf = linear_resample_time(df, dt=delta, with_timespan=True, keep_nan=True)
        for name, template in params:
            ldf.loc[ldf[TIMESPAN_ID].isnull(), [name]] = zero
            cumsum = ldf[name].cumsum()
            for target in mins:
                n = (target * 60) // delta
                diff = cumsum.diff(periods=n).dropna()
                if present(diff, name):
                    stats[template % target] = diff.max() / n
    except Exception as e:
        log.warning(f'No Max Mean stats: {e}')
    return stats


def max_med_stats(df, params=((HEART_RATE, MAX_MED_HR_M),), mins=None, delta=10, gap=0.01):
    stats, mins = {}, mins or MAX_MINUTES
    try:
        ldf_all = linear_resample_time(df, dt=delta, with_timespan=False, add_time=False)
        ldf_all.interpolate('nearest')
        ldf_tstamp = ldf_all.loc[ldf_all[TIMESPAN_ID].isin(df[TIMESPAN_ID].unique())].copy()
        ldf_tstamp.loc[:, 'gap'] = ldf_tstamp.index.astype(np.int64) / 1e9
        ldf_tstamp.loc[:, 'gap'] = ldf_tstamp['gap'].diff()
        log.debug(f'Largest gap is {ldf_tstamp["gap"].max()}s')
        for target in mins:
            n = target * 60 // delta
            log.debug(f'Target {target}m is {n} samples (delta {delta}s)')
            splits, remain = [], ldf_all.copy()
            max_gap = max(gap * target * 60, 1.5 * delta)
            for after in ldf_tstamp.index[ldf_tstamp['gap'] > max_gap].tolist():
                before = ldf_tstamp.index[ldf_tstamp.index.get_loc(after)-1]
                splits.append(remain.loc[:before])
                remain = remain.loc[after:]
            splits.append(remain)
            log.debug(f'Split data into {len(splits)} sections for {target}m with max gap of {max_gap}s')
            for name, template in params:
                stat_name = template % target
                for split in splits:
                    split['med'] = split[name].rolling(n).median()
                    if present(split, 'med'):
                        max_med = split['med'].dropna().max()
                        if stat_name in stats:
                            stats[stat_name] = max(stats[stat_name], max_med)
                        else:
                            stats[stat_name] = max_med
    except Exception as e:
        log.warning(f'No Max/Med stats: {e}')
    return stats


def direction_stats(df):
    stats = {}
    try:
        if all(name in df.columns for name in (SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y)):
            df = df.dropna(subset=[SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y]).copy()
            if not df.empty:
                x0, y0 = df.iloc[0][SPHERICAL_MERCATOR_X], df.iloc[0][SPHERICAL_MERCATOR_Y]
                df.loc[:, 'dx'] = df[SPHERICAL_MERCATOR_X] - x0
                df.loc[:, 'dy'] = df[SPHERICAL_MERCATOR_Y] - y0
                # average position
                dx, dy = df['dx'].mean(), df['dy'].mean()
                x1, y1, d = x0 + dx, y0 + dy, sqrt(dx ** 2 + dy ** 2)
                theta = atan2(dy, dx)
                # change coords to centred on average position and perp / parallel to line to start
                df.loc[:, 'dx'] = df[SPHERICAL_MERCATOR_X] - x1
                df.loc[:, 'dy'] = df[SPHERICAL_MERCATOR_Y] - y1
                df.loc[:, 'u'] = df['dx'] * cos(theta) + df['dy'] * sin(theta)
                df.loc[:, 'v'] = df['dy'] * cos(theta) - df['dx'] * sin(theta)
                # convert from angle anti-clock from x axis to bearing
                stats[DIRECTION] = 90 - 180 * theta / pi
                stats[ASPECT_RATIO] = df['v'].std() / df['u'].std()
    except Exception as e:
        log.warning(f'No Direction stats: {e}')
    return stats


# if __name__ == '__main__':
#     from . import *
#     date = '2017-02-07 07:18:50'
#     s = session('-v5')
#     df = activity_statistics(s, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y,
#                              local_time=date, activity_group_name='Bike', with_timespan=True)
#     print(direction_stats(df))




from enum import IntEnum
from json import dumps

from sqlalchemy import Column, Integer

from ..support import Base
from ..types import Cls, Json, Sort


class PipelineType(IntEnum):

    STATISTIC = 0
    DIARY = 1
    ACTIVITY = 2
    MONITOR = 3


class Pipeline(Base):

    __tablename__ = 'pipeline'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False)
    cls = Column(Cls, nullable=False)
    args = Column(Json, nullable=False, server_default=dumps(()))
    kargs = Column(Json, nullable=False, server_default=dumps({}))
    sort = Column(Sort)

    @classmethod
    def all(cls, s, type, like=None, id=None):
        q = s.query(Pipeline).filter(Pipeline.type == type)
        if like:
            q = q.filter(Pipeline.cls.like(like))
        if id:
            q = q.filter(Pipeline.id == id)
        pipelines = q.order_by(Pipeline.sort).all()
        if not pipelines:
            msg = 'No pipelines configured for type %s' % PipelineType(type).name
            if like:
                msg += f' like {like}'
            if id:
                msg += f' with ID={id}'
            raise Exception(msg)
        yield from pipelines




from abc import abstractmethod
from logging import getLogger

from ..pipeline import MultiProcPipeline, UniProcPipeline, LoaderMixin
from ... import FatalException
from ...fit.format.read import filtered_records
from ...fit.profile.profile import read_fit
from ...lib.date import to_time
from ...lib.io import filter_modified_files, update_scan
from ...lib.log import log_current_exception
from ...squeal import Timestamp

log = getLogger(__name__)


class AbortImport(Exception):
    pass


class AbortImportButMarkScanned(AbortImport):
    pass


class FitReaderMixin(LoaderMixin):

    def __init__(self, *args, paths=None, **kargs):
        self.paths = paths
        super().__init__(*args, **kargs)

    def _delete(self, s):
        # forcing does two things:
        # first, it by-passes the checks on last-scan date and duplicates
        # second, we delete any overlapping activities on load (when we know the times)
        # (without force, overlapping activities trigger an error)
        pass

    def _missing(self, s):
        return filter_modified_files(s, self.paths, self.owner_out, self.force)

    def _run_one(self, s, path):
        try:
            self._read(s, path)
            update_scan(s, path, self.owner_out)
        except AbortImportButMarkScanned as e:
            log.warning(f'Could not process {path} (scanned)')
            # log_current_exception()
            update_scan(s, path, self.owner_out)
        except FatalException:
            raise
        except Exception as e:
            log.warning(f'Could not process {path} (ignored)')
            log_current_exception()

    def _read(self, s, path):
        key, data = self._read_data(s, path)
        s.commit()
        with Timestamp(owner=self.owner_out, key=key).on_success(log, s):
            loader = self._get_loader(s)
            self._load_data(s, loader, data)
            loader.load()

    def _read_fit_file(self, path, *options):
        types, messages, records = filtered_records(read_fit(path))
        return [record.as_dict(*options)
                for _, _, record in sorted(records,
                                           key=lambda r: r[2].timestamp if r[2].timestamp else to_time(0.0))]

    def _first(self, path, records, *names):
        return self.__assert_contained(path, records, names, 0)

    def _last(self, path, records, *names):
        return self.__assert_contained(path, records, names, -1)

    def __assert_contained(self, path, records, names, index):
        try:
            return [record for record in records if record.name in names][index]
        except IndexError:
            log.debug(f'No {names} entry(s) in {path}')
            raise AbortImportButMarkScanned()

    @abstractmethod
    def _read_data(self, s, path):
        raise NotImplementedError()

    @abstractmethod
    def _load_data(self, s, loader, data):
        raise NotImplementedError()


class MultiProcFitReader(FitReaderMixin, MultiProcPipeline):

    def _args(self, missing, start, finish):
        paths = ' '.join(repr(path) for path in missing[start:finish+1])  # quote names
        log.info(f'Starting worker for {missing[start]} - {missing[finish]}')
        return paths


class UniProcFitReader(FitReaderMixin, UniProcPipeline):

    def _base_command(self):
        raise Exception('UniProc does not support workers')

    def _args(self, missing, start, finish):
        raise Exception('UniProc does not support workers')




from logging import getLogger
from os.path import splitext, basename

from pygeotile.point import Point
from sqlalchemy.sql.functions import count

from ..names import LATITUDE, LONGITUDE, M, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, ELEVATION, RAW_ELEVATION
from ..read import AbortImport, MultiProcFitReader, AbortImportButMarkScanned
from ...commands.args import ACTIVITIES, WORKER, FAST, mm, FORCE
from ...fit.format.records import fix_degrees, merge_duplicates, no_bad_values
from ...lib.date import to_time
from ...sortem.bilinear import bilinear_elevation_from_constant
from ...squeal.database import Timestamp, StatisticJournalText
from ...squeal.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan
from ...squeal.tables.statistic import StatisticJournalFloat, STATISTIC_JOURNAL_CLASSES
from ...squeal.utils import add

log = getLogger(__name__)


# duplicate data in
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24715592701_tap-sync-18690-cc1dd93225119215a1ea87c584a974ce.fit
# /home/andrew/archive/fit/batch/DI_CONNECT/DI-Connect-Fitness/UploadedFiles_0-_Part1/andrew@acooke.org_24718989709_tap-sync-18690-effaaaffdd06b9419991471bd92d53d5.fit

class ActivityReader(MultiProcFitReader):

    def __init__(self, *args, constants=None, sport_to_activity=None, record_to_db=None, **kargs):
        self.constants = constants
        self.sport_to_activity = self._assert('sport_to_activity', sport_to_activity)
        self.record_to_db = [(field, name, units, STATISTIC_JOURNAL_CLASSES[type])
                             for field, (name, units, type)
                             in self._assert('record_to_db', record_to_db).items()]
        self.add_elevation = not any(name == ELEVATION for (field, name, units, type) in self.record_to_db)
        super().__init__(*args, **kargs)

    def _base_command(self):
        if self.constants:
            constants = ' '.join(f'-D "{constant}"' for constant in self.constants) + ' -- '
        else:
            constants = ''
        return f'{{ch2}} -v0 -l {{log}} -f {self._db.path} {ACTIVITIES} {mm(WORKER)} {self.id} ' \
            f'{mm(FAST)} {mm(FORCE) if self.force else ""} {constants}'

    def _startup(self, s):
        super()._startup(s)
        self.__oracle = bilinear_elevation_from_constant(log, s)

    def _read_data(self, s, path):
        log.info('Reading activity data from %s' % path)
        records = self._read_fit_file(path, merge_duplicates, fix_degrees, no_bad_values)
        ajournal, activity_group, first_timestamp = self._create_activity(s, path, records)
        self._load_constants(s, ajournal)
        return ajournal.id, (ajournal, activity_group, first_timestamp, path, records)

    def __read_sport(self, path, records):
        try:
            return self._first(path, records, 'sport').value.sport.lower()
        except AbortImportButMarkScanned:
            # alternative for some garmin devices (florian)
            return self._first(path, records, 'session').value.sport.lower()

    def _create_activity(self, s, path, records):
        first_timestamp = self._first(path, records, 'event', 'record').value.timestamp
        last_timestamp = self._last(path, records, 'event', 'record').value.timestamp
        log.debug(f'Time range: {first_timestamp.timestamp()} - {last_timestamp.timestamp()}')
        sport = self.__read_sport(path, records)
        activity_group = self._activity_group(s, path, sport)
        log.info(f'{activity_group} from {sport}')
        if self.force:
            self._delete_journals(s, activity_group, first_timestamp, last_timestamp)
        else:
            self._check_journals(s, activity_group, first_timestamp, last_timestamp)
        ajournal = add(s, ActivityJournal(activity_group=activity_group,
                                          start=first_timestamp, finish=first_timestamp,  # will be over-written later
                                          fit_file=path, name=splitext(basename(path))[0]))
        return ajournal, activity_group, first_timestamp

    def _activity_group(self, s, path, sport):
        if sport in self.sport_to_activity:
            return self._lookup_activity_group(s, self.sport_to_activity[sport])
        else:
            log.warning('Unrecognised sport: "%s" in %s' % (sport, path))
            raise AbortImport()

    def _lookup_activity_group(self, s, name):
        activity_group = s.query(ActivityGroup).filter(ActivityGroup.name == name).one_or_none()
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                log.info('Available activity group:')
                for activity_group in activities:
                    log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                log.error('No activity groups defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    def _overlapping_journals(self, s, what, activity_group, first_timestamp, last_timestamp):
        return s.query(what). \
                filter(ActivityJournal.activity_group == activity_group,
                       ActivityJournal.start < last_timestamp,
                       ActivityJournal.finish >= first_timestamp)

    def _delete_journals(self, s, activity_group, first_timestamp, last_timestamp):
        for journal in self._overlapping_journals(s, ActivityJournal, activity_group,
                                                  first_timestamp, last_timestamp).all():
            Timestamp.clear(s, owner=self.owner_out, key=journal.id)
            log.debug(f'Deleting {journal}')
            s.delete(journal)
        s.commit()

    def _check_journals(self, s, activity_group, first_timestamp, last_timestamp):
        if self._overlapping_journals(s, count(ActivityJournal.id), activity_group,
                                      first_timestamp, last_timestamp).scalar():
            log.warning(f'Overlapping activities for {first_timestamp} - {last_timestamp}')
            raise AbortImportButMarkScanned()

    def _load_constants(self, s, ajournal):
        if self.constants:
            for constant in self.constants:
                name, value = constant.split('=', 1)
                log.debug(f'Setting {name}={value}')
                StatisticJournalText.add(log, s, name, None, None, self.owner_out, ajournal.activity_group,
                                         ajournal, value, ajournal.start)

    def _load_data(self, s, loader, data):

        ajournal, activity_group, first_timestamp, path, records = data
        timespan, warned, logged, last_timestamp = None, 0, 0, to_time(0.0)
        log.debug(f'Loading {self.record_to_db}')

        def is_event(record, *types):
            event = False
            if record.name == 'event':
                event = record.value.event == 'timer' and record.value.event_type in types
                # log.debug(f'Event: {event} for {types} ({record})')
                if event: log.debug(f'{types} at {record.timestamp}')
            return event

        have_timespan = any(is_event(record, 'start') for record in records)
        only_records = list(filter(lambda x: x.name == 'record', records))
        final_timestamp = only_records[-1].timestamp
        if not have_timespan:
            first_timestamp = only_records[0].timestamp
            log.warning('Experimental handling of data without timespans')
            timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                               start=first_timestamp))

        for record in records:

            if have_timespan and is_event(record, 'start'):
                if timespan:
                    log.warning('Ignoring start with no corresponding stop (possible lost data?)')
                else:
                    timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                                       start=record.value.timestamp,
                                                       finish=record.value.timestamp))

            elif record.name == 'record':
                if record.value.timestamp > last_timestamp:
                    lat, lon, timestamp = None, None, record.value.timestamp
                    # customizable loader
                    for field, name, units, type in self.record_to_db:
                        value = record.data.get(field, None)
                        if logged < 3:
                            log.debug(f'{name} = {value}')
                        if value is not None:
                            value = value[0][0]
                            loader.add(name, units, None, activity_group, ajournal, value, timestamp, type)
                            if name == LATITUDE:
                                lat = value
                            elif name == LONGITUDE:
                                lon = value
                    logged += 1
                    # values derived from lat/lon
                    if lat is not None and lon is not None:
                        x, y = Point.from_latitude_longitude(lat, lon).meters
                        loader.add(SPHERICAL_MERCATOR_X, M, None, activity_group, ajournal, x, timestamp,
                                   StatisticJournalFloat)
                        loader.add(SPHERICAL_MERCATOR_Y, M, None, activity_group, ajournal, y, timestamp,
                                   StatisticJournalFloat)
                        if self.add_elevation:
                            elevation = self.__oracle.elevation(lat, lon)
                            if elevation:
                                loader.add(RAW_ELEVATION, M, None, activity_group, ajournal, elevation, timestamp,
                                           StatisticJournalFloat)
                else:
                    log.warning('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                (path, record.value.timestamp))
                last_timestamp = record.value.timestamp
                if not have_timespan:
                    ajournal.finish = record.value.timestamp

            elif have_timespan and is_event(record, 'stop_all', 'stop'):
                if timespan:
                    timespan.finish = record.value.timestamp
                    ajournal.finish = record.value.timestamp
                    timespan = None
                else:
                    log.debug('Ignoring stop with no corresponding start (possible lost data?)')

        if timespan:
            log.warning('Cleaning up dangling timespan')
            timespan.finish = final_timestamp




from re import escape

from sqlalchemy import select, or_, and_

from .frame import _tables
from ..lib.peg import transform, choice, pattern, sequence, Recursive, drop, exhaustive, single
from ..squeal import ActivityJournal, StatisticName, StatisticJournalType


def pat(regexp, trans=None):
    parser = pattern(r'\s*' + regexp + r'\s*')
    if trans:
        parser = transform(parser, trans)
    return parser


def lit(text, trans=None):
    return pat('(' + escape(text) + ')', trans=trans)


def unexp(values):
    try:
        val, exp = values
        return [float(val) * 10 ** int(exp)]
    except:
        return [float(values[0])]


def to_tuple(l):
    return [tuple(l)]


def join(a, sep, b):
    return transform(sequence(a, sep, b), to_tuple)


def flip(l):
    value, op, name = l[0]
    return [(name, {'=': '=', '!=': '!=', '<': '>', '>': '<', '>=': '<=', '<=': '>='}[op], value)]


name = pat(r'(\w(?:\s*\w)*)')

number = choice(pat(r'(-?\d+)', int), pat(r'([-+]?\d*\.?\d+)(?:[eE]([-+]?\d+))?', unexp))

string = choice(pat(r'"((?:[^"\\]|\\.)*)"'), pat(r"'((?:[^'\\]|\\.)*)'"))

value = choice(number, string)

operator = choice(*[lit(cmp) for cmp in ['=', '!=', '<', '>', '<=', '>=']])

comparison = choice(join(name, operator, value), transform(join(value, operator, name), flip))

term = Recursive()

and_comparison = Recursive()
and_comparison.calls(choice(term, join(term, lit('&'), term)))

or_comparison = Recursive()
or_comparison.calls(choice(and_comparison, join(and_comparison, lit('|'), or_comparison)))

parens = transform(sequence(drop(lit('(')), or_comparison, drop(lit(')'))))

term.calls(choice(comparison, parens))

constraint = single(exhaustive(choice(or_comparison, parens)))


def constrained_activities(s, query):
    ast = constraint(query)[0]
    q = build_activity_query(s, ast)
    return list(q)


def build_activity_query(s, ast):
    t = _tables()
    constraints = build_constraints(s, t, ast)
    return s.query(ActivityJournal).filter(ActivityJournal.id.in_(constraints))


def build_constraints(s, t, ast):
    l, op, r = ast
    if op in '&|':
        lcte = build_constraints(s, t, l)
        rcte = build_constraints(s, t, r)
        return build_join(t, op, lcte, rcte)
    else:
        return build_comparisons(s, t, ast)


def build_join(t, op, lcte, rcte):
    if op == '|':
        return select([t.aj.c.id]).where(or_(t.aj.c.id.in_(lcte), t.aj.c.id.in_(rcte)))
    else:
        return select([t.aj.c.id]).where(and_(t.aj.c.id.in_(lcte), t.aj.c.id.in_(rcte)))


def build_comparisons(s, t, ast):
    name, op, value = ast
    comparisons = [build_comparison(t, statistic_name, op, value)
                   for statistic_name in s.query(StatisticName).filter(StatisticName.name == name).all()]
    if len(comparisons) == 1:
        return comparisons[0]
    else:
        return join_comparisons(comparisons)


def build_comparison(t, statistic_name, op, value):
    if statistic_name.statistic_journal_type == StatisticJournalType.INTEGER:
        table = t.sji
    elif statistic_name.statistic_journal_type == StatisticJournalType.FLOAT:
        table = t.sjf
    else:
        table = t.sjs
    q = select([t.sj.c.source_id]).select_from(t.sj.join(table)).where(t.sj.c.statistic_name_id == statistic_name.id)
    attr = {'=': '__eq__', '!=': '__ne__', '>': '__gt__', '>=': '__ge__', '<': '__lt__', '<=': '__le__'}[op]
    q = q.where(getattr(table.c.value, attr)(value))
    return q


def join_comparisons(comparisons):
    raise NotImplemented()




import datetime as dt
from logging import getLogger

import numpy as np
import pandas as pd
from sqlalchemy import desc, and_, or_, distinct, func, select
from sqlalchemy.sql.functions import count

from ..load import StatisticJournalLoader
from ..names import HEART_RATE, BPM, STEPS, STEPS_UNITS, CUMULATIVE_STEPS, _new, TIME, SOURCE
from ..read import AbortImportButMarkScanned, AbortImport, MultiProcFitReader
from ...commands.args import MONITOR, WORKER, FAST, mm, FORCE
from ...data.frame import _tables
from ...fit.format.records import fix_degrees, unpack_single_bytes, merge_duplicates
from ...lib.date import time_to_local_date, format_time
from ...squeal.database import StatisticJournalType
from ...squeal.tables.monitor import MonitorJournal
from ...squeal.tables.statistic import StatisticJournalInteger, StatisticName, StatisticJournal
from ...squeal.utils import add

log = getLogger(__name__)
ACTIVITY_TYPE_ATTR = 'activity_type'
HEART_RATE_ATTR = 'heart_rate'
MONITORING_ATTR = 'monitoring'
MONITORING_INFO_ATTR = 'monitoring_info'
STEPS_ATTR = 'steps'


def missing_dates(s):
    # we don't try to be perfect here.  the idea is that it's called to get the latest
    # updates, rather than fill in all gaps (do the bulk download thing for that).
    # we also don't try to get fractional data.
    # and as for timezones... we just assume garmin uses the local timezone.
    latest = s.query(MonitorJournal).order_by(desc(MonitorJournal.start)).limit(1).one_or_none()
    if latest is None:
        log.warning('No existing monitor data - ' +
                    'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
        return
    # find the mid-point to avoid any problems with timezones and edge cases
    # (these files don't span more than a day)
    seconds = (latest.finish - latest.start).total_seconds() / 2
    start = time_to_local_date(latest.start + dt.timedelta(seconds=seconds))
    finish = dt.date.today()
    days = (finish - start).days
    if days > 14:
        raise Exception('Too many days (%d) - ' % days +
                        'do a bulk download instead: https://www.garmin.com/en-US/account/datamanagement/')
    if days:
        while start < finish:
            yield start
            start += dt.timedelta(days=1)
    else:
        log.warning('No dates to download')


class MonitorLoader(StatisticJournalLoader):

    def _preload(self):
        dummy = super()._preload()
        try:
            for name in self._s.query(StatisticName). \
                    filter(StatisticName.name == CUMULATIVE_STEPS,
                           StatisticName.owner == self._owner).all():
                n = self._s.query(count(StatisticJournal.id)). \
                    filter(StatisticJournal.statistic_name == name,
                           StatisticJournal.time >= self.start,
                           StatisticJournal.time <= self.finish).scalar()
                if n and self.start and self.finish:
                    log.warning(f'Deleting {n} overlapping {CUMULATIVE_STEPS}')
                    self._s.query(StatisticJournal). \
                        filter(StatisticJournal.statistic_name == name,
                               StatisticJournal.time >= self.start,
                               StatisticJournal.time <= self.finish).delete()
        except:
            log.debug('Failed to clean database')
            self._s.rollback()
            raise
        return dummy

    def _resolve_duplicate(self, name, instance, prev):
        log.warning(f'Using max of duplicate values at {instance.time} for {name} ({instance.value}/{prev.value})')
        prev.value = max(prev.value, instance.value)


NEW_STEPS = _new(STEPS)


class MonitorReader(MultiProcFitReader):

    def _get_loader(self, s, **kargs):
        if 'owner' not in kargs:
            kargs['owner'] = self.owner_out
        return MonitorLoader(s, **kargs)

    def _base_command(self):
        return f'{{ch2}} -v0 -l {{log}} -f {self._db.path} {MONITOR} {mm(WORKER)} {self.id} {mm(FAST)} {mm(FORCE) if self.force else ""}'

    def _delete_contained(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(MonitorJournal.start >= start,
                       MonitorJournal.finish <= finish).all():
            log.warning(f'Replacing {mjournal.fit_file} ({mjournal.start} - {mjournal.finish}) with data from '
                        f'{path} ({start} - {finish}')
            s.delete(mjournal)

    def _check_inside(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(or_(and_(MonitorJournal.start <= start, MonitorJournal.finish > finish),
                           and_(MonitorJournal.start < start, MonitorJournal.finish >= finish))).all():
            log.warning('%s already includes data from %s for %s - %s' % (mjournal, path, start, finish))
            raise AbortImportButMarkScanned()

    def _check_overlap(self, s, start, finish, path):
        for mjournal in s.query(MonitorJournal). \
                filter(or_(and_(MonitorJournal.start < start,
                                MonitorJournal.finish > start, MonitorJournal.finish < finish),
                           and_(MonitorJournal.finish > finish,
                                MonitorJournal.start > start, MonitorJournal.start < finish))).all():
            log.warning('%s overlaps data from %s for %s - %s' % (mjournal, path, start, finish))
            raise AbortImport()

    def _delete_previous(self, s, start, finish, path):
        self._check_inside(s, start, finish, path)
        self._check_overlap(s, start, finish, path)
        self._delete_contained(s, start, finish, path)
        s.commit()

    def _read_data(self, s, path):

        records = self._read_fit_file(path, merge_duplicates, fix_degrees, unpack_single_bytes)

        first_timestamp = self._first(path, records, MONITORING_INFO_ATTR).timestamp
        last_timestamp = self._last(path, records, MONITORING_ATTR).timestamp
        if first_timestamp == last_timestamp:
            log.debug('File %s is empty (no timespan)' % path)
            raise AbortImportButMarkScanned()
        if not first_timestamp:
            raise Exception('Missing timestamp in %s' % path)

        log.info(f'Importing monitor data from {path} '
                 f'for {format_time(first_timestamp)} - {format_time(last_timestamp)}')
        self._delete_previous(s, first_timestamp, last_timestamp, path)
        mjournal = add(s, MonitorJournal(start=first_timestamp, fit_file=path, finish=last_timestamp))

        return mjournal.id, (first_timestamp, last_timestamp, mjournal, records)

    def _load_data(self, s, loader, data):
        first_timestamp, last_timestamp, mjournal, records = data
        for record in records:
            if HEART_RATE_ATTR in record.data:
                loader.add(HEART_RATE, BPM, None, None, mjournal, record.data[HEART_RATE_ATTR][0][0],
                           record.timestamp, StatisticJournalInteger)
            if STEPS_ATTR in record.data:
                for (activity, steps) in zip(record.data[ACTIVITY_TYPE_ATTR][0], record.data[STEPS_ATTR][0]):
                    loader.add(CUMULATIVE_STEPS, STEPS_UNITS, None, activity, mjournal, steps,
                               record.timestamp, StatisticJournalInteger)

    def _shutdown(self, s):
        super()._shutdown(s)
        if not self.worker:
            for activity in self._step_activities(s):
                df = self._read_diff(s, activity)
                df = self._calculate_diff(df)
                self._write_diff(s, df, activity)

    def _step_activities(self, s):
        return [row[0] for row in s.query(distinct(StatisticName.constraint)).
            filter(StatisticName.name == CUMULATIVE_STEPS,
                   StatisticName.owner == self.owner_out).all()]

    def _read_diff(self, s, activity):
        t = _tables()
        qs = select([t.sj.c.time.label("time"), t.sji.c.value.label("steps")]). \
            select_from(t.sj.join(t.sn).join(t.sji)). \
            where(and_(t.sn.c.name == STEPS, t.sn.c.constraint == activity,
                       t.sn.c.owner == self.owner_out)).alias("steps")
        q = select([t.sj.c.time.label(TIME), t.sj.c.source_id.label(SOURCE), t.sji.c.value.label(CUMULATIVE_STEPS),
                    qs.c.steps.label(STEPS)]). \
            select_from(t.sj.join(t.sn).join(t.sji).outerjoin(qs, t.sj.c.time == qs.c.time)). \
            where(and_(t.sn.c.name == CUMULATIVE_STEPS, t.sn.c.constraint == activity,
                       t.sn.c.owner == self.owner_out)). \
            order_by(t.sj.c.time)
        # log.debug(q)
        df = pd.read_sql_query(sql=q, con=s.connection(), index_col=TIME)
        return df

    def _calculate_diff(self, df):
        df[NEW_STEPS] = df[CUMULATIVE_STEPS].diff()
        df.loc[df[NEW_STEPS] < 0, NEW_STEPS] = df[CUMULATIVE_STEPS]
        df.loc[df[NEW_STEPS].isna(), NEW_STEPS] = df[CUMULATIVE_STEPS]
        return df

    def _write_diff(self, s, df, activity):
        steps = StatisticName.add_if_missing(log, s, STEPS, StatisticJournalType.INTEGER, STEPS_UNITS, None,
                                             self.owner_out, activity)
        times = df.loc[(df[NEW_STEPS] != df[STEPS]) & ~df[STEPS].isna()].index.astype(np.int64) / 1e9
        if len(times):
            n = s.query(func.count(StatisticJournal.id)). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).scalar()
            log.warning(f'Deleting {n} {STEPS}/{activity} entries')
            s.query(StatisticJournal.id). \
                filter(StatisticJournal.time.in_(times),
                       StatisticJournal.statistic_name == steps).delete(synchronize_session=False)
        loader = StatisticJournalLoader(s, owner=self.owner_out)
        for time, row in df.loc[(df[NEW_STEPS] != df[STEPS]) & ~df[NEW_STEPS].isna()].iterrows():
            loader.add(STEPS, STEPS_UNITS, None, activity, row[SOURCE], int(row[NEW_STEPS]),
                       time, StatisticJournalInteger)
        loader.load()




from logging import getLogger
from subprocess import run
from tempfile import NamedTemporaryFile
from unittest import TestCase

import sqlalchemy.sql.functions as func

from ch2 import monitor
from ch2.commands.args import bootstrap_file, m, V, DEV, mm, FAST
from ch2.config.default import default
from ch2.lib.date import to_time, local_date_to_time
from ch2.squeal.tables.monitor import MonitorJournal
from ch2.squeal.tables.pipeline import PipelineType
from ch2.squeal.tables.statistic import StatisticJournal, StatisticName
from ch2.stoats.calculate.monitor import MonitorCalculator
from ch2.stoats.names import REST_HR, DAILY_STEPS
from ch2.stoats.pipeline import run_pipeline

log = getLogger(__name__)


class TestMonitor(TestCase):

    def test_monitor(self):
        with NamedTemporaryFile() as f:
            args, db = bootstrap_file(f, m(V), '5')
            bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            args, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                      'monitor', mm(FAST), 'data/test/source/personal/25822184777.fit')
            monitor(args, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(db, PipelineType.STATISTIC, force=True, start='2018-01-01')
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                n = s.query(func.count(StatisticJournal.id)).scalar()
                # self.assertEqual(n, 111)
                self.assertEqual(n, 108)  # why?
                mjournal = s.query(MonitorJournal).one()
                self.assertNotEqual(mjournal.start, mjournal.finish)

    def test_values(self):
        with NamedTemporaryFile() as f:
            bootstrap_file(f, m(V), '5')
            bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            for file in ('24696157869', '24696160481', '24696163486'):
                args, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                          'monitor', mm(FAST),
                                          'data/test/source/personal/andrew@acooke.org_%s.fit' % file)
                monitor(args, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(db, PipelineType.STATISTIC, force=True, start='2018-01-01', n_cpu=1)
            run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                mjournals = s.query(MonitorJournal).order_by(MonitorJournal.start).all()
                assert mjournals[2].start == to_time('2018-09-06 15:06:00'), mjournals[2].start
                print(mjournals[2].fit_file)
                # steps
                summary = s.query(StatisticJournal).join(StatisticName). \
                    filter(StatisticJournal.time >= local_date_to_time('2018-09-06'),
                           StatisticJournal.time < local_date_to_time('2018-09-07'),
                           StatisticName.owner == MonitorCalculator,
                           StatisticName.name == DAILY_STEPS).one()
                # connect has 12757 for this date,
                self.assertEqual(summary.value, 12757)
                # heart rate
                summary = s.query(StatisticJournal).join(StatisticName). \
                    filter(StatisticJournal.time >= local_date_to_time('2018-09-06'),
                           StatisticJournal.time < local_date_to_time('2018-09-07'),
                           StatisticName.owner == MonitorCalculator,
                           StatisticName.name == REST_HR).one()
                self.assertEqual(summary.value, 45)

    FILES = ('25505915679', '25519562859', '25519565531', '25532154264', '25539076032', '25542112328')

    def generic_bug(self, files):
        with NamedTemporaryFile() as f:
            args, db = bootstrap_file(f, m(V), '5')
            bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            for file in files:
                args, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                               'monitor', mm(FAST),
                                               'data/test/source/personal/andrew@acooke.org_%s.fit' % file)
                monitor(args, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(db, PipelineType.STATISTIC, force=True, start='2018-01-01', n_cpu=1)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                # steps
                summary = s.query(StatisticJournal).join(StatisticName). \
                    filter(StatisticJournal.time >= local_date_to_time('2018-10-07'),
                           StatisticJournal.time < local_date_to_time('2018-10-08'),
                           StatisticName.owner == MonitorCalculator,
                           StatisticName.name == DAILY_STEPS).one()
                # connect has 3031 for this date.
                self.assertEqual(summary.value, 3031)

    def test_bug(self):
        self.generic_bug(sorted(self.FILES))

    def test_bug_reversed(self):
        self.generic_bug(sorted(self.FILES, reverse=True))

    # issue 6
    def test_empty_data(self):
        with NamedTemporaryFile() as f:
            args, db = bootstrap_file(f, m(V), '5')
            bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)
            args, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                      'monitor', mm(FAST), 'data/test/source/other/37140810636.fit')
            monitor(args, db)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            run_pipeline(db, PipelineType.STATISTIC, n_cpu=1)
            # run('sqlite3 %s ".dump"' % f.name, shell=True)
            with db.session_context() as s:
                n = s.query(func.count(StatisticJournal.id)).scalar()
                self.assertEqual(n, 19)
                mjournal = s.query(MonitorJournal).one()
                self.assertNotEqual(mjournal.start, mjournal.finish)
